import csv
import math
from io import StringIO

import numpy as np
import pandas as pd
from wbfdm.models import InstrumentPrice
from xlrd import xldate_as_datetime

from wbportfolio.models import Product, Register, Trade

from .sylk import SYLK
from .utils import assemble_transaction_reference, get_portfolio_id


def convert_string_to_number(string):
    try:
        return float(string.replace(" ", "").replace(",", ""))
    except ValueError:
        return 0.0


def remove_not_needed_columns(df: pd.DataFrame):
    del df["NORDER"]
    del df["NORDER_EXTERN"]
    del df["REGISTER_ID1"]
    del df["ISIN1"]
    del df["AMOUNT"]
    del df["QUANTITY"]
    del df["AMOUNT_EST_EUR"]
    del df["CODE_OPERATION"]
    del df["TRADE_DATE"]
    del df["VALUE_DATE"]
    del df["REGISTER_DEAL_NAME"]
    del df["REGISTER_ID2"]


def convert_dates(df: pd.DataFrame):
    df["trade_date"] = df["TRADE_DATE"].apply(lambda x: xldate_as_datetime(x, datemode=0).date())
    df["trade_date"] = df["trade_date"].apply(lambda x: x.strftime("%Y-%m-%d"))
    df["value_date"] = df["VALUE_DATE"].apply(lambda x: xldate_as_datetime(x, datemode=0).date())
    df["value_date"] = df["value_date"].apply(lambda x: x.strftime("%Y-%m-%d"))


def map_products_by_isin(df: pd.DataFrame):
    product_mapping = {
        product["isin"]: product["id"]
        for product in Product.objects.filter(isin__in=df["ISIN1"].unique()).values("id", "isin")
    }
    df["underlying_instrument"] = df["ISIN1"].apply(lambda x: product_mapping[x])
    df["portfolio"] = df["underlying_instrument"].apply(lambda x: get_portfolio_id(x))


def map_registers_by_references(df: pd.DataFrame):
    register_mapping = {
        register["register_reference"]: register["id"]
        for register in Register.objects.filter(register_reference__in=df["REGISTER_ID1"].unique()).values(
            "id", "register_reference"
        )
    }
    df["register"] = df["REGISTER_ID1"].apply(lambda x: register_mapping[str(x)])


def get_price(x):
    try:
        estimated_price = float(
            InstrumentPrice.objects.filter(
                instrument=x["underlying_instrument"], calculated=False, date__lte=x["trade_date"]
            )
            .latest("date")
            .net_value
        )
    except InstrumentPrice.DoesNotExist:
        try:
            estimated_price = Product.objects.get(id=x["underlying_instrument"]).issue_price
        except Product.DoesNotExist:
            estimated_price = 100
    return estimated_price


def get_shares(x):
    multiplier = 1 if x["CODE_OPERATION"] == 1 else -1
    if x["QUANTITY"] and not math.isnan(x["QUANTITY"]):
        return round(x["QUANTITY"] * multiplier, 4)

    return round((x["AMOUNT"] * multiplier) / x["price"], 4)


def parse(import_source):
    data = list()

    sylk_handler = SYLK()
    for line in [_line.decode("cp1252") for _line in import_source.file.open("rb").readlines()]:
        sylk_handler.parseline(line)

    buffer = StringIO()
    csvwriter = csv.writer(buffer, quotechar="'", delimiter=";", lineterminator="\n", quoting=csv.QUOTE_ALL)
    for line in sylk_handler.stream_rows():
        csvwriter.writerow(line)

    buffer.seek(0)
    content = buffer.read().replace('""', "")
    df = pd.read_csv(StringIO(content), sep=";", quotechar="'", usecols=[0, 4, 5, 6, 7, 12, 13, 15, 16, 17, 18, 22])
    if not df.empty:
        df["pending"] = True

        df.loc[df["CODE_OPERATION"] == 6, "pending"] = False
        transfers = df.loc[df["CODE_OPERATION"] == 6, :].copy()
        transfers["CODE_OPERATION"] = 2
        transfers["QUANTITY"] = -transfers["QUANTITY"]
        transfers["TRANSFER_REGISTER"] = transfers["REGISTER_ID1"].astype("int").astype("str")
        transfers["REGISTER_ID1"] = transfers["REGISTER_ID2"].astype("int").astype("str")
        transfers["REGISTER_DEAL_NAME"] = transfers["REGISTER_ID1"].apply(
            lambda x: Register.objects.get(register_reference=x).register_name_1
        )

        df = pd.concat([df, transfers], axis=0)

        if "TRANSFER_REGISTER" in df:
            df["TRANSFER_REGISTER"] = df["TRANSFER_REGISTER"].replace(np.nan, "", regex=True)

        convert_dates(df)
        df["NORDER_EXTERN"] = df["NORDER_EXTERN"].replace(np.nan, "", regex=True)
        map_products_by_isin(df)
        df["price"] = df.apply(get_price, axis=1)
        df["shares"] = df.apply(get_shares, axis=1)
        df["transaction_subtype"] = df.shares.apply(
            lambda x: Trade.Type.REDEMPTION.value if x < 0 else Trade.Type.SUBSCRIPTION.value
        )

        df["bank"] = df["REGISTER_DEAL_NAME"]
        df["external_id_alternative"] = df["NORDER"]
        df["register__register_reference"] = df["REGISTER_ID1"]
        df["external_id"] = df.apply(assemble_transaction_reference, axis=1)

        if "TRANSFER_REGISTER" in df:
            del df["TRANSFER_REGISTER"]
        remove_not_needed_columns(df)
        # Convert df to list of dicts
        data = df.rename(columns={"trade_date": "transaction_date"}).to_dict("records")
    return {"data": data}



from django.contrib import admin

from wbportfolio.models.transactions.claim import Claim


class ClaimTabularInline(admin.TabularInline):
    extra = 0
    model = Claim
    list_display = ["reference_id", "product", "shares", "nominal_value", "claimant"]
    autocomplete_fields = ["account", "product", "claimant"]


@admin.register(Claim)
class ClaimModelAdmin(admin.ModelAdmin):
    list_display = ["date", "reference_id", "product", "account", "shares", "nominal_value"]
    autocomplete_fields = ["account", "product", "claimant"]
    raw_id_fields = ["trade", "account", "product", "claimant", "creator"]



import pandas as pd

from wbportfolio.models import Trade

from .utils import _get_underlying_instrument, file_name_parse_isin


def _parse_trade_type(type):
    choices = {
        "Decrease": Trade.Type.DECREASE,
        "Increase": Trade.Type.INCREASE,
        "Rebalancing": Trade.Type.REBALANCE,
        "Buy": Trade.Type.BUY,
        "Sell": Trade.Type.SELL,
    }

    return choices[type]


def parse(import_source):
    data = list()
    df = pd.read_csv(import_source.file, sep=";")
    if not df.empty:
        # Parse the Parts of the filename into the different parts
        parts = file_name_parse_isin(import_source.file.name)

        # Get the valuation date and investment from the parts list
        product_data = parts["product"]

        # Iterate through the CSV File and parse the data into a list
        df["underlying_instrument"] = df[["BLOOMBERG CODE", "NAME", "QUOTED_CRNCY"]].apply(
            lambda x: _get_underlying_instrument(*x), axis=1
        )
        df["exchange"] = df.underlying_instrument.apply(lambda x: x.get("exchange", None))
        columns_map = {
            "underlying_instrument": "underlying_instrument",
            "TRADE DATE": "transaction_date",
            "EXECUTED QTY": "shares",
            "EXECUTED PRICE": "price",
            "FX RATE": "currency_fx_rate",
            "NET PRICE(Stock crncy)": "price",
            "PRICE(USD)": "price_gross",
            "TRADE TYPE": "transaction_subtype",
            "QUOTED_CRNCY": "currency__key",
        }
        df = df.rename(columns=columns_map)
        float_fields = ["shares", "price", "currency_fx_rate", "price", "price_gross"]
        for field in float_fields:
            if field in df:
                df[field] = df[field].apply(
                    lambda x: float(x.replace("-", "").replace(" ", "") if isinstance(x, str) else x)
                )

        df["price"] = df.apply(
            lambda x: x["price"] * x["QUOTITY/ADJ. FACTOR"] if "QUOTITY/ADJ. FACTOR" in x else x["price"], axis=1
        )
        df.transaction_date = pd.to_datetime(df.transaction_date, dayfirst=True)
        df.transaction_date = df.transaction_date.apply(lambda x: x.strftime("%Y-%m-%d"))
        df["transaction_subtype"] = df["transaction_subtype"].apply(lambda x: _parse_trade_type(x))
        df = df.drop(columns=df.columns.difference(columns_map.values()))
        df["portfolio"] = [{"instrument_type": "product", **product_data}] * df.shape[0]
        df["bank"] = "Natixis Cash Transfer"
        data = df.to_dict("records")
    return {"data": data}



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

OrderProposalMenuItem = MenuItem(
    label="Order Proposals",
    endpoint="wbportfolio:orderproposal-list",
    endpoint_get_parameters={"waiting_for_input": True},
    permission=ItemPermission(
        permissions=["wbportfolio.view_orderproposal"], method=lambda request: is_internal_user(request.user)
    ),
)



import logging
import re

from wbcore.contrib.currency.models import Currency
from wbcore.contrib.directory.models import Company

logger = logging.getLogger("pms")


def get_currency_and_assets_under_management(
    aum_string, currency_mapping, default_currency, multiplier_mapping, default_multiplier
):
    currency = default_currency
    multiplier = default_multiplier

    regex_number = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", aum_string)

    if regex_number:
        regex_number = regex_number[0]
    else:
        return None, None

    aum_string = aum_string.lower()

    for mapping in currency_mapping.keys():
        if mapping in aum_string:
            currency = currency_mapping[mapping]
            aum_string = aum_string.replace(mapping, "")

    for mapping in multiplier_mapping.keys():
        if mapping in aum_string:
            multiplier = multiplier_mapping[mapping]
            aum_string = aum_string.replace(mapping, "")

    return float(regex_number) * multiplier, currency


def assign_aum():
    default_currency = Currency.objects.get(key="USD")

    currency_mapping = {
        "chf": Currency.objects.get(key="CHF"),
        "eur": Currency.objects.get(key="EUR"),
        "€": Currency.objects.get(key="EUR"),
        "usd": Currency.objects.get(key="USD"),
        "$": Currency.objects.get(key="USD"),
    }

    multiplier_mapping = {
        "billion": 1000000000,
        "bn": 1000000000,
        "b.": 1000000000,
        "b": 1000000000,
        "million": 1000000,
        "mio": 1000000,
        "mm": 1000000,
        "mn": 1000000,
        "m": 1000000,
    }

    for company in Company.objects.all():
        assets_under_management, currency = get_currency_and_assets_under_management(
            aum_string=company.assets_under_management,
            currency_mapping=currency_mapping,
            default_currency=default_currency,
            multiplier_mapping=multiplier_mapping,
            default_multiplier=1000000,
        )

        if not hasattr(company, "portfolio_data"):
            company.save()

        portfolio_data = company.portfolio_data
        portfolio_data.assets_under_management = assets_under_management
        portfolio_data.assets_under_management_currency = currency
        try:
            portfolio_data.save()
        except Exception as e:
            logger.error(
                f"while we try to save the customer portfolio aum data for {company}, we encounter the error: {e}"
            )



import datetime
import logging
import re

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import convert_string_to_number

logger = logging.getLogger("importers.parsers.jp_morgan.strategy")


def file_name_parse(file_name):
    dates = re.findall("([0-9]{8})", file_name)

    if len(dates) != 1:
        raise ValueError("Not exactly 1 date found in the filename")
    return {"valuation_date": datetime.datetime.strptime(dates[0], "%Y%m%d").date()}


def manually_create_100_position(parent_strategies, valuation_date):
    data = []
    from wbportfolio.models import Index, Product

    for strategy_ticker in parent_strategies:
        if index := Index.objects.filter(ticker=strategy_ticker).first():
            for product in Product.objects.filter(ticker=strategy_ticker):
                valuations = product.valuations.filter(date__lte=valuation_date)
                last_price = 0
                if valuations.exists():
                    last_price = float(valuations.latest("date").net_value)
                data.append(
                    {
                        "underlying_quote": index.id,
                        "portfolio": {"instrument_type": "product", "id": product.id},
                        "currency__key": index.currency.key,
                        "initial_currency_fx_rate": 1.0,
                        "weighting": 1.0,
                        "is_estimated": True,  # this position is not a real position, it is created by the importer.
                        "initial_price": last_price,
                        "date": valuation_date.strftime("%Y-%m-%d"),
                    }
                )
    return data


def parse(import_source):
    # Load file into a CSV DictReader

    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=",")
    df = df.replace([np.inf, -np.inf, np.nan], None)

    # Parse the Parts of the filename into the different parts
    parts = file_name_parse(import_source.file.name)

    # Get the valuation date from the parts list
    valuation_date = parts["valuation_date"]

    # Iterate through the CSV File and parse the data into a list
    data = list()
    parents_strategies = set()
    for strategy_data in df.to_dict("records"):
        bbg_tickers = strategy_data["BBG Ticker"].split(" ")
        exchange = None
        if len(bbg_tickers) == 2:
            ticker = bbg_tickers[0]
            instrument_type = bbg_tickers[1]
        elif len(bbg_tickers) == 3:
            ticker = bbg_tickers[0]
            exchange = bbg_tickers[1]
            instrument_type = bbg_tickers[2]

        strategy = strategy_data["Strategy Ticker"].replace("Index", "").strip()
        strategy_currency_key = strategy_data["Strategy CCY"]

        position_currency_key = strategy_data["Position CCY"]

        isin = strategy_data["Position ISIN"]
        name = strategy_data["Position Description"]
        initial_price = convert_string_to_number(strategy_data["Prices"])
        initial_currency_fx_rate = convert_string_to_number(strategy_data["Fx Rates"])
        if exchange:
            exchange = {"bbg_exchange_codes": exchange}
        try:
            weighting = convert_string_to_number(strategy_data["Weight In Percent"].replace("%", "")) / 100
        except Exception:
            weighting = 0.0
        underlying_quote = {
            "ticker": ticker,
            "exchange": exchange,
            "isin": isin,
            "name": name,
            "currency__key": position_currency_key,
            "instrument_type": instrument_type.lower(),
        }
        if isin:
            underlying_quote["isin"] = isin
        data.append(
            {
                "underlying_quote": underlying_quote,
                "portfolio": {
                    "instrument_type": "index",
                    "ticker": strategy,
                    "currency__key": strategy_currency_key,
                },
                "exchange": exchange,
                "is_estimated": False,
                "currency__key": position_currency_key,
                "initial_currency_fx_rate": initial_currency_fx_rate,
                "weighting": weighting,
                "initial_price": initial_price,
                "date": valuation_date.strftime("%Y-%m-%d"),
            }
        )
        parents_strategies.add(strategy)
    manual_data = manually_create_100_position(parents_strategies, valuation_date)
    return {"data": data + manual_data}



import codecs
import csv
import datetime

from wbportfolio.import_export.utils import convert_string_to_number

from .utils import file_name_parse_isin


def parse(import_source):
    # Load file into a CSV DictReader
    csv_file = import_source.file.open(mode="rt")
    csv_file_type = type(csv_file.read())

    # If the file is a byte string, then we need to convert it.
    if csv_file_type is bytes:
        csv_file = codecs.iterdecode(csv_file, "utf-8")

    # Read file into a CSV Dict Reader
    csv_reader = csv.DictReader(csv_file, delimiter=";")

    # Parse the Parts of the filename into the different parts
    parts = file_name_parse_isin(import_source.file.name)

    # Get the valuation date and investment from the parts list
    product_data = parts["product"]

    # Iterate through the CSV File and parse the data into a list
    data = list()

    for valuation in csv_reader:
        date = datetime.datetime.strptime(valuation["Date"], "%d/%m/%Y").date()
        if date.weekday() not in [5, 6]:
            data.append(
                {
                    "instrument": {"instrument_type": "product", **product_data},
                    "date": date.strftime("%Y-%m-%d"),
                    "net_value": round(convert_string_to_number(valuation["Index Value in%"]), 6),
                    "gross_value": round(convert_string_to_number(valuation["Index gross in%"]), 6),
                    "calculated": False,
                }
            )

    csv_file.close()
    return {"data": data}



# Import necessary modules
from datetime import date, timedelta
from decimal import Decimal
from unittest.mock import call, patch

import pytest
from django.db.models import Sum
from faker import Faker
from pandas._libs.tslibs.offsets import BDay, BusinessMonthEnd

from wbportfolio.models import Order, OrderProposal, Portfolio, RebalancingModel
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position

fake = Faker()


# Mark tests to use Django's database
@pytest.mark.django_db
class TestOrderProposal:
    def test_init(self, order_proposal):
        assert order_proposal.id is not None

    # Test that the checked object is correctly set to the portfolio
    def test_checked_object(self, order_proposal):
        """
        Verify that the checked object is the portfolio associated with the order proposal.
        """
        assert order_proposal.checked_object == order_proposal.portfolio

    # Test that the evaluation date matches the trade date
    def test_check_evaluation_date(self, order_proposal):
        """
        Ensure the evaluation date is the same as the trade date.
        """
        assert order_proposal.check_evaluation_date == order_proposal.trade_date

    # Test the validated trading service functionality
    def test_validated_trading_service(self, order_proposal, asset_position_factory, order_factory):
        """
        Validate that the effective and target portfolios are correctly calculated.
        """
        effective_date = (order_proposal.trade_date - BDay(1)).date()

        # Create asset positions for testing
        a1 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.3")
        )
        a2 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.7")
        )

        # Create orders for testing
        o1 = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal("0.05"),
            portfolio=order_proposal.portfolio,
            underlying_instrument=a1.underlying_quote,
        )
        o2 = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal("-0.05"),
            portfolio=order_proposal.portfolio,
            underlying_instrument=a2.underlying_quote,
        )

        r1 = o1.price / a1.initial_price - Decimal("1")
        r2 = o2.price / a2.initial_price - Decimal("1")
        p_return = a1.weighting * (Decimal("1") + r1) + a2.weighting * (Decimal("1") + r2)
        # Get the validated trading service
        trades = order_proposal.validated_trading_service.trades_batch.trades_map
        t1 = trades[a1.underlying_quote.id]
        t2 = trades[a2.underlying_quote.id]

        # Assert effective and target portfolios are as expected
        assert t1.previous_weight == a1.weighting
        assert t2.previous_weight == a2.weighting
        assert t1.target_weight == pytest.approx(
            a1.weighting * ((r1 + Decimal("1")) / p_return) + o1.weighting, abs=Decimal("1e-8")
        )
        assert t2.target_weight == pytest.approx(
            a2.weighting * ((r2 + Decimal("1")) / p_return) + o2.weighting, abs=Decimal("1e-8")
        )

    # Test the calculation of the last effective date
    def test_last_effective_date(self, order_proposal, asset_position_factory):
        """
        Verify the last effective date is correctly determined based on asset positions.
        """
        # Without any positions, it should be the day before the trade date
        assert (
            order_proposal.last_effective_date == (order_proposal.trade_date - BDay(1)).date()
        ), "Last effective date without position should be t-1"

        # Create an asset position before the trade date
        a1 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=(order_proposal.trade_date - BDay(5)).date()
        )
        a_noise = asset_position_factory.create(portfolio=order_proposal.portfolio, date=order_proposal.trade_date)  # noqa

        # The last effective date should still be the day before the trade date due to caching
        assert (
            order_proposal.last_effective_date == (order_proposal.trade_date - BDay(1)).date()
        ), "last effective date is cached, so it won't change as is"

        # Reset the cache property to recalculate
        del order_proposal.last_effective_date

        # Now it should be the date of the latest position before the trade date
        assert (
            order_proposal.last_effective_date == a1.date
        ), "last effective date is the latest position strictly lower than trade date"

    # Test finding the previous order proposal
    def test_previous_order_proposal(self, order_proposal_factory):
        """
        Ensure the previous order proposal is correctly identified as the last approved proposal before the current one.
        """
        tp = order_proposal_factory.create()
        tp_previous_submit = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.PENDING, trade_date=(tp.trade_date - BDay(1)).date()
        )
        tp_previous_approve = order_proposal_factory.create(
            portfolio=tp.portfolio, status=OrderProposal.Status.APPLIED, trade_date=(tp.trade_date - BDay(2)).date()
        )
        tp_next_approve = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.APPLIED, trade_date=(tp.trade_date + BDay(1)).date()
        )

        # The previous valid order proposal should be the approved one strictly before the current proposal
        assert (
            tp.previous_order_proposal == tp_previous_approve
        ), "the previous valid order proposal is the strictly before and approved order proposal"

    # Test finding the next order proposal
    def test_next_order_proposal(self, order_proposal_factory):
        """
        Verify the next order proposal is correctly identified as the first approved proposal after the current one.
        """
        tp = order_proposal_factory.create()
        tp_previous_approve = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.APPLIED, trade_date=(tp.trade_date - BDay(1)).date()
        )
        tp_next_submit = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.PENDING, trade_date=(tp.trade_date + BDay(1)).date()
        )
        tp_next_approve = order_proposal_factory.create(
            portfolio=tp.portfolio, status=OrderProposal.Status.APPLIED, trade_date=(tp.trade_date + BDay(2)).date()
        )

        # The next valid order proposal should be the approved one strictly after the current proposal
        assert (
            tp.next_order_proposal == tp_next_approve
        ), "the next valid order proposal is the strictly after and approved order proposal"

    # Test getting the default target portfolio
    def test__get_default_target_portfolio(self, order_proposal, asset_position_factory):
        """
        Ensure the default target portfolio is set to the effective portfolio from the day before the trade date.
        """
        effective_date = (order_proposal.trade_date - BDay(1)).date()

        # Create asset positions for testing
        a1 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.3")
        )
        a2 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.7")
        )
        asset_position_factory.create(portfolio=order_proposal.portfolio, date=order_proposal.trade_date)  # noise

        # The default target portfolio should match the effective portfolio
        assert order_proposal._get_default_target_portfolio().to_dict() == {
            a1.underlying_quote.id: a1.weighting,
            a2.underlying_quote.id: a2.weighting,
        }

    # Test getting the default target portfolio with a rebalancing model
    @patch.object(RebalancingModel, "get_target_portfolio")
    def test__get_default_target_portfolio_with_rebalancer_model(self, mock_fct, order_proposal, rebalancer_factory):
        """
        Verify that the target portfolio is correctly obtained from a rebalancing model.
        """
        # Expected target portfolio from the rebalancing model
        expected_target_portfolio = PortfolioDTO(
            positions=(Position(underlying_instrument=1, weighting=Decimal(1), date=order_proposal.trade_date),)
        )
        mock_fct.return_value = expected_target_portfolio

        # Create a rebalancer for testing
        rebalancer = rebalancer_factory.create(
            portfolio=order_proposal.portfolio, parameters={"rebalancer_parameter": "A"}
        )
        order_proposal.rebalancing_model = rebalancer.rebalancing_model
        order_proposal.save()

        # Additional keyword arguments for the rebalancing model
        extra_kwargs = {"test": "test"}

        # Combine rebalancer parameters with extra keyword arguments
        expected_kwargs = rebalancer.parameters
        expected_kwargs.update(extra_kwargs)

        # Assert the target portfolio matches the expected output from the rebalancing model
        assert (
            order_proposal._get_default_target_portfolio(**extra_kwargs) == expected_target_portfolio
        ), "We expect the target portfolio to be whatever is returned by the rebalancer model"
        mock_fct.assert_called_once_with(
            order_proposal.portfolio, order_proposal.trade_date, order_proposal.last_effective_date, **expected_kwargs
        )

    # Test normalizing orders
    def test_normalize_orders(self, order_proposal, order_factory):
        """
        Ensure orders are normalized to sum up to 1, handling quantization errors.
        """
        # Create orders for testing
        t1 = order_factory.create(
            order_proposal=order_proposal,
            portfolio=order_proposal.portfolio,
            weighting=Decimal(0.2),
        )
        t2 = order_factory.create(
            order_proposal=order_proposal,
            portfolio=order_proposal.portfolio,
            weighting=Decimal(0.26),
        )
        t3 = order_factory.create(
            order_proposal=order_proposal,
            portfolio=order_proposal.portfolio,
            weighting=Decimal(0.14),
        )

        # Normalize orders
        order_proposal.normalize_orders()

        # Refresh orders from the database
        t1.refresh_from_db()
        t2.refresh_from_db()
        t3.refresh_from_db()

        # Expected normalized weights
        normalized_t1_weight = Decimal("0.33333333")
        normalized_t2_weight = Decimal("0.43333333")
        normalized_t3_weight = Decimal("0.23333333")

        # Calculate quantization error
        quantize_error = Decimal(1) - (normalized_t1_weight + normalized_t2_weight + normalized_t3_weight)

        # Assert quantization error exists and weights are normalized correctly
        assert quantize_error
        assert t1.weighting == normalized_t1_weight
        assert t2.weighting == normalized_t2_weight + quantize_error  # Add quantize error to the largest position
        assert t3.weighting == normalized_t3_weight

    # Test resetting orders
    def test_reset_orders(
        self, order_proposal, instrument_factory, cash_factory, instrument_price_factory, asset_position_factory
    ):
        """
        Verify orders are correctly reset based on effective and target portfolios.
        """
        cash = cash_factory.create()
        effective_date = order_proposal.last_effective_date

        # Create instruments for testing
        i1 = instrument_factory.create(currency=order_proposal.portfolio.currency)
        i2 = instrument_factory.create(currency=order_proposal.portfolio.currency)
        i3 = instrument_factory.create(currency=order_proposal.portfolio.currency)
        # Build initial effective portfolio constituting only from two positions of i1 and i2
        asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, underlying_instrument=i1, weighting=Decimal("0.7")
        )
        asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, underlying_instrument=i2, weighting=Decimal("0.3")
        )
        p1 = instrument_price_factory.create(instrument=i1, date=effective_date)
        p2 = instrument_price_factory.create(instrument=i2, date=effective_date)
        p3 = instrument_price_factory.create(instrument=i3, date=effective_date)

        # build the target portfolio
        target_portfolio = PortfolioDTO(
            [
                Position(
                    underlying_instrument=i2.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.4"),
                    price=float(p2.net_value),
                ),
                Position(
                    underlying_instrument=i3.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.6"),
                    price=float(p3.net_value),
                ),
            ]
        )

        # Reset orders
        order_proposal.reset_orders(target_portfolio=target_portfolio)

        # Get orders for each instrument
        t1 = order_proposal.orders.get(underlying_instrument=i1)
        t2 = order_proposal.orders.get(underlying_instrument=i2)
        t3 = order_proposal.orders.get(underlying_instrument=i3)

        # Assert trade weights are correctly reset
        assert t1.weighting == Decimal("-0.7")
        assert t2.weighting == Decimal("0.1")
        assert t3.weighting == Decimal("0.6")

        # build the target portfolio
        new_target_portfolio = PortfolioDTO(
            [
                Position(
                    underlying_instrument=i1.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.2"),
                    price=float(p1.net_value),
                ),
                Position(
                    underlying_instrument=i2.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.3"),
                    price=float(p2.net_value),
                ),
                Position(
                    underlying_instrument=i3.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.5"),
                    price=float(p3.net_value),
                ),
            ]
        )

        order_proposal.reset_orders(target_portfolio=new_target_portfolio)
        # Refetch the orders for each instrument
        t1.refresh_from_db()
        t2.refresh_from_db()
        t3.refresh_from_db()
        # Assert existing trade weights are correctly updated
        assert t1.weighting == Decimal("-0.5")
        assert t2.weighting == Decimal("0")
        assert t3.weighting == Decimal("0.5")

        # assert cash position creates a proper order
        # build the target portfolio
        target_portfolio_with_cash = PortfolioDTO(
            [
                Position(
                    underlying_instrument=i1.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.5"),
                    price=float(p1.net_value),
                ),
                Position(
                    underlying_instrument=cash.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.5"),
                    price=1.0,
                ),
            ]
        )
        order_proposal.reset_orders(target_portfolio=target_portfolio_with_cash)

        # Assert existing trade weights are correctly updated
        assert order_proposal.orders.get(underlying_instrument=i1).weighting == Decimal("-0.2")
        assert order_proposal.orders.get(underlying_instrument=i2).weighting == Decimal("-0.3")
        assert order_proposal.orders.get(underlying_instrument=cash).weighting == Decimal("0.5")

    def test_reset_orders_remove_invalid_orders(self, order_proposal, order_factory):
        # create a invalid trade and its price
        invalid_trade = order_factory.create(order_proposal=order_proposal, weighting=Decimal(0))

        # create a valid trade and its price
        valid_trade = order_factory.create(order_proposal=order_proposal, weighting=Decimal(1))
        order_proposal.reset_orders()
        assert order_proposal.orders.get(underlying_instrument=valid_trade.underlying_instrument).weighting == Decimal(
            "1"
        )
        with pytest.raises(Order.DoesNotExist):
            order_proposal.orders.get(underlying_instrument=invalid_trade.underlying_instrument)

    # Test replaying order proposals
    @patch.object(Portfolio, "drift_weights")
    def test_replay(self, mock_fct, order_proposal_factory):
        """
        Ensure replaying order proposals correctly calls drift_weights for each period.
        """
        mock_fct.return_value = iter([])

        # Create approved order proposals for testing
        tp0 = order_proposal_factory.create(status=OrderProposal.Status.APPLIED)
        tp1 = order_proposal_factory.create(
            portfolio=tp0.portfolio,
            status=OrderProposal.Status.APPLIED,
            trade_date=(tp0.trade_date + BusinessMonthEnd(1)).date(),
        )
        tp2 = order_proposal_factory.create(
            portfolio=tp0.portfolio,
            status=OrderProposal.Status.APPLIED,
            trade_date=(tp1.trade_date + BusinessMonthEnd(1)).date(),
        )

        # Replay order proposals
        tp0.replay()

        # Expected calls to drift_weights
        expected_calls = [
            call(tp0.trade_date, tp1.trade_date - timedelta(days=1), stop_at_rebalancing=True),
            call(tp1.trade_date, tp2.trade_date - timedelta(days=1), stop_at_rebalancing=True),
            call(tp2.trade_date, date.today(), stop_at_rebalancing=True),
        ]

        # Assert drift_weights was called as expected
        mock_fct.assert_has_calls(expected_calls)

        # Test stopping replay on a non-approved proposal
        tp1.status = OrderProposal.Status.FAILED
        tp1.save()
        expected_calls = [call(tp0.trade_date, tp1.trade_date - timedelta(days=1), stop_at_rebalancing=True)]
        mock_fct.assert_has_calls(expected_calls)

    # Test estimating shares for a trade
    @patch.object(Portfolio, "get_total_asset_value")
    def test_get_estimated_shares(
        self, mock_fct, order_proposal, order_factory, instrument_price_factory, instrument_factory
    ):
        """
        Verify shares estimation based on trade weighting and instrument price.
        """
        portfolio = order_proposal.portfolio
        instrument = instrument_factory.create(currency=portfolio.currency)
        underlying_quote_price = instrument_price_factory.create(instrument=instrument, date=order_proposal.trade_date)
        mock_fct.return_value = Decimal(1_000_000)  # 1 million cash
        trade = order_factory.create(
            order_proposal=order_proposal,
            value_date=order_proposal.trade_date,
            portfolio=portfolio,
            underlying_instrument=instrument,
        )
        trade.refresh_from_db()

        # Assert estimated shares are correctly calculated
        assert (
            order_proposal.get_estimated_shares(
                trade.weighting, trade.underlying_instrument, underlying_quote_price.net_value
            )
            == Decimal(1_000_000) * trade.weighting / underlying_quote_price.net_value
        )

    @patch.object(Portfolio, "get_total_asset_value")
    def test_get_estimated_target_cash(self, mock_fct, order_proposal, order_factory, cash_factory):
        order_proposal.portfolio.only_weighting = False
        order_proposal.portfolio.save()
        mock_fct.return_value = Decimal(1_000_000)  # 1 million cash
        cash = cash_factory.create(currency=order_proposal.portfolio.currency)
        order_factory.create(  # equity trade
            order_proposal=order_proposal,
            value_date=order_proposal.trade_date,
            portfolio=order_proposal.portfolio,
            weighting=Decimal("0.7"),
        )
        order_factory.create(  # cash trade
            order_proposal=order_proposal,
            value_date=order_proposal.trade_date,
            portfolio=order_proposal.portfolio,
            underlying_instrument=cash,
            weighting=Decimal("0.2"),
        )
        target_cash_position = order_proposal.get_estimated_target_cash()
        assert target_cash_position.weighting == Decimal("0.3")
        assert target_cash_position.initial_shares == Decimal(1_000_000) * Decimal("0.3")

    def test_order_proposal_update_inception_date(self, order_proposal_factory, portfolio, instrument_factory):
        # Check that if we create a prior order proposal, the instrument inception date is updated accordingly
        instrument = instrument_factory.create(inception_date=None)
        instrument.portfolios.add(portfolio)
        tp = order_proposal_factory.create(portfolio=portfolio)
        instrument.refresh_from_db()
        assert instrument.inception_date == tp.trade_date

        tp2 = order_proposal_factory.create(portfolio=portfolio, trade_date=(tp.trade_date - BDay(1)).date())
        instrument.refresh_from_db()
        assert instrument.inception_date == tp2.trade_date

    def test_get_round_lot_size(self, order_proposal, instrument):
        # without a round lot size, we expect no normalization of shares
        assert order_proposal.get_round_lot_size(Decimal("66"), instrument) == Decimal("66")
        instrument.round_lot_size = 100
        instrument.save()

        # if instrument has a round lot size different than 1, we expect different behavior based on whether shares is positive or negative
        assert order_proposal.get_round_lot_size(Decimal(66.0), instrument) == Decimal("100")
        assert order_proposal.get_round_lot_size(Decimal(-66.0), instrument) == Decimal(-66.0)
        assert order_proposal.get_round_lot_size(Decimal(-120), instrument) == Decimal(-200)

        # exchange can disable rounding based on the lot size
        instrument.exchange.apply_round_lot_size = False
        instrument.exchange.save()
        assert order_proposal.get_round_lot_size(Decimal("66"), instrument) == Decimal("66")

    @patch.object(Portfolio, "get_total_asset_value")
    def test_submit_round_lot_size(self, mock_fct, order_proposal, instrument_price_factory, order_factory):
        initial_shares = Decimal("70")
        price = instrument_price_factory.create(date=order_proposal.trade_date)
        net_value = round(price.net_value, 4)
        portfolio_value = initial_shares * net_value
        mock_fct.return_value = portfolio_value

        order_proposal.portfolio.only_weighting = False
        order_proposal.portfolio.save()
        instrument = price.instrument
        instrument.round_lot_size = 100
        instrument.save()
        trade = order_factory.create(
            shares=initial_shares,
            order_proposal=order_proposal,
            weighting=Decimal("1.0"),
            underlying_instrument=price.instrument,
            price=net_value,
        )
        warnings = order_proposal.submit()
        order_proposal.save()
        assert (
            len(warnings) == 1
        )  # ensure that submit returns a warning concerning the rounded trade based on the lot size
        trade.refresh_from_db()
        assert trade.shares == 100  # we expect the share to be transformed from 70 to 100 (lot size of 100)

    @patch.object(Portfolio, "get_total_asset_value")
    def test_submit_round_fractional_shares(self, mock_fct, instrument_price_factory, order_proposal, order_factory):
        initial_shares = Decimal("5.6")
        price = instrument_price_factory.create(date=order_proposal.trade_date)
        net_value = round(price.net_value, 4)
        portfolio_value = initial_shares * net_value
        mock_fct.return_value = portfolio_value

        order_proposal.portfolio.only_weighting = False
        order_proposal.portfolio.save()
        trade = order_factory.create(
            shares=Decimal("5.6"),
            order_proposal=order_proposal,
            weighting=Decimal("1.0"),
            underlying_instrument=price.instrument,
            price=net_value,
        )
        order_proposal.submit()
        order_proposal.save()
        trade.refresh_from_db()
        assert trade.shares == 6  # we expect the fractional share to be rounded
        assert trade.weighting == round((trade.shares * net_value) / portfolio_value, 8)
        assert trade.weighting == round(
            Decimal("1") + ((Decimal("6") - initial_shares) * net_value) / portfolio_value, 8
        )  # we expect the weighting to be updated accrodingly

    def test_ex_post(
        self, instrument_factory, asset_position_factory, instrument_price_factory, order_proposal_factory, portfolio
    ):
        """
        Tests the ex-post rebalancing mechanism of a portfolio with two instruments.
        Verifies that weights are correctly recalculated after submitting and approving a order proposal.
        """

        # --- Create instruments ---
        msft = instrument_factory.create(currency=portfolio.currency)
        apple = instrument_factory.create(currency=portfolio.currency)

        # --- Key dates ---
        d1 = date(2025, 6, 24)
        d2 = date(2025, 6, 25)
        d3 = date(2025, 6, 26)
        d4 = date(2025, 6, 27)

        # --- Create MSFT prices ---
        msft_p1 = instrument_price_factory.create(instrument=msft, date=d1, net_value=Decimal("10"))
        msft_p2 = instrument_price_factory.create(instrument=msft, date=d2, net_value=Decimal("8"))
        msft_p3 = instrument_price_factory.create(instrument=msft, date=d3, net_value=Decimal("12"))
        msft_p4 = instrument_price_factory.create(instrument=msft, date=d4, net_value=Decimal("15"))  # noqa

        # Calculate MSFT returns between dates
        msft_r2 = msft_p2.net_value / msft_p1.net_value - Decimal("1")  # noqa
        msft_r3 = msft_p3.net_value / msft_p2.net_value - Decimal("1")

        # --- Create Apple prices (stable) ---
        apple_p1 = instrument_price_factory.create(instrument=apple, date=d1, net_value=Decimal("100"))
        apple_p2 = instrument_price_factory.create(instrument=apple, date=d2, net_value=Decimal("100"))
        apple_p3 = instrument_price_factory.create(instrument=apple, date=d3, net_value=Decimal("100"))
        apple_p4 = instrument_price_factory.create(instrument=apple, date=d4, net_value=Decimal("100"))  # noqa

        # Apple returns (always 0 since price is stable)
        apple_r2 = apple_p2.net_value / apple_p1.net_value - Decimal("1")  # noqa
        apple_r3 = apple_p3.net_value / apple_p2.net_value - Decimal("1")

        # --- Create positions on d2 ---
        msft_a2 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_quote=msft,
            date=d2,
            initial_shares=10,
            weighting=Decimal("0.44"),
        )
        apple_a2 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_quote=apple,
            date=d2,
            initial_shares=1,
            weighting=Decimal("0.56"),
        )

        # Check that initial weights sum to 1
        total_weight_d2 = msft_a2.weighting + apple_a2.weighting
        assert total_weight_d2 == pytest.approx(Decimal("1.0"), abs=Decimal("1e-6"))

        # --- Calculate total portfolio return between d2 and d3 ---
        portfolio_r3 = msft_a2.weighting * (Decimal("1.0") + msft_r3) + apple_a2.weighting * (
            Decimal("1.0") + apple_r3
        )

        # --- Create positions on d3 with weights adjusted for returns ---

        # Check that weights on d2 sum to 1
        total_weight_d2 = msft_a2.weighting + apple_a2.weighting
        assert total_weight_d2 == pytest.approx(Decimal("1.0"), abs=Decimal("1e-6"))

        # --- Create a order proposal on d3 ---
        order_proposal = order_proposal_factory.create(portfolio=portfolio, trade_date=d3)
        order_proposal.reset_orders()
        # Retrieve orders for each instrument
        orders = order_proposal.get_orders()
        trade_msft = orders.get(underlying_instrument=msft)
        trade_apple = orders.get(underlying_instrument=apple)
        # Check that trade weights are initially zero
        assert trade_msft.weighting == Decimal("0")
        assert trade_apple.weighting == Decimal("0")

        msft_drifted = msft_a2.weighting * (Decimal("1.0") + msft_r3) / portfolio_r3
        apple_drifted = apple_a2.weighting * (Decimal("1.0") + apple_r3) / portfolio_r3
        # --- Adjust trade weights to target 50% each ---
        target_weight = Decimal("0.5")
        trade_msft.weighting = target_weight - msft_drifted
        trade_msft.save()

        trade_apple.weighting = target_weight - apple_drifted
        trade_apple.save()
        orders = order_proposal.get_orders()
        trade_msft = orders.get(underlying_instrument=msft)
        trade_apple = orders.get(underlying_instrument=apple)

        # --- Check drift factors and effective weights ---
        assert trade_msft.daily_return == pytest.approx(msft_r3, abs=Decimal("1e-6"))
        assert trade_apple.daily_return == pytest.approx(apple_r3, abs=Decimal("1e-6"))

        assert trade_msft._effective_weight == pytest.approx(msft_drifted, abs=Decimal("1e-6"))
        assert trade_apple._effective_weight == pytest.approx(apple_drifted, abs=Decimal("1e-6"))

        # Check that the target weight is the sum of drifted weight and adjustment
        assert trade_msft._target_weight == pytest.approx(target_weight, abs=Decimal("1e-6"))
        assert trade_apple._target_weight == pytest.approx(target_weight, abs=Decimal("1e-6"))

        # --- Submit and approve the order proposal ---
        order_proposal.submit()
        order_proposal.save()
        order_proposal.approve()
        order_proposal.apply()
        order_proposal.save()

        # Final check that weights have been updated to 50%
        assert order_proposal.portfolio.assets.get(underlying_instrument=msft).weighting == pytest.approx(
            target_weight, abs=Decimal("1e-6")
        )
        assert order_proposal.portfolio.assets.get(underlying_instrument=apple).weighting == pytest.approx(
            target_weight, abs=Decimal("1e-6")
        )

    def test_replay_reset_draft_order_proposal(
        self, instrument_factory, instrument_price_factory, order_factory, order_proposal_factory
    ):
        instrument = instrument_factory.create()
        order_proposal = order_proposal_factory.create(trade_date=date.today() - BDay(2))
        instrument_price_factory.create(instrument=instrument, date=date.today() - BDay(2))
        instrument_price_factory.create(instrument=instrument, date=date.today() - BDay(1))
        instrument_price_factory.create(instrument=instrument, date=date.today())
        trade = order_factory.create(
            underlying_instrument=instrument,
            order_proposal=order_proposal,
            weighting=1,
        )
        order_proposal.submit()
        order_proposal.approve()
        order_proposal.apply(replay=False)
        order_proposal.save()

        draft_tp = order_proposal_factory.create(portfolio=order_proposal.portfolio, trade_date=date.today() - BDay(1))
        assert not Order.objects.filter(order_proposal=draft_tp).exists()

        order_proposal.replay()

        assert Order.objects.filter(order_proposal=draft_tp).count() == 1
        assert Order.objects.get(
            order_proposal=draft_tp, underlying_instrument=trade.underlying_instrument
        ).weighting == Decimal("0")

    def test_order_submit_bellow_minimum_allowed_order_value(self, order_factory):
        order = order_factory.create(price=Decimal(1), weighting=Decimal(1), shares=Decimal(999))
        order.submit()
        order.save()
        assert order.shares == Decimal(999)
        assert order.weighting == Decimal(1)

        order.order_proposal.min_order_value = Decimal(1000)
        order.order_proposal.save()

        order.submit()
        order.save()
        assert order.shares == Decimal(0)
        assert order.weighting == Decimal(0)

    def test_order_submit_bellow_minimum_weighting(self, order_factory, order_proposal):
        o1 = order_factory.create(order_proposal=order_proposal, price=Decimal(1), weighting=Decimal("0.8"))
        o2 = order_factory.create(order_proposal=order_proposal, price=Decimal(1), weighting=Decimal("0.2"))
        order_proposal.submit()
        order_proposal.save()

        o1.refresh_from_db()
        o2.refresh_from_db()
        assert o1.weighting == Decimal("0.8")
        assert o2.weighting == Decimal("0.2")

        order_proposal.min_weighting = Decimal("0.21")
        order_proposal.backtodraft()
        order_proposal.submit()
        order_proposal.save()

        o1.refresh_from_db()
        o2.refresh_from_db()
        assert o1.weighting == Decimal("0.8")
        assert o2.weighting == Decimal("0")

        order_proposal.approve()
        order_proposal.apply()
        order_proposal.save()
        assert order_proposal.portfolio.assets.get(
            date=order_proposal.trade_date, underlying_quote=o1.underlying_instrument
        ).weighting == Decimal("0.8")
        assert order_proposal.portfolio.assets.get(
            date=order_proposal.trade_date, underlying_quote=order_proposal.cash_component
        ).weighting == Decimal("0.2")

    def test_reset_order_use_desired_target_weight(self, order_proposal, order_factory):
        order1 = order_factory.create(
            order_proposal=order_proposal, weighting=Decimal("0.5"), desired_target_weight=Decimal("0.7")
        )
        order2 = order_factory.create(
            order_proposal=order_proposal, weighting=Decimal("0.5"), desired_target_weight=Decimal("0.3")
        )
        order_proposal.submit()
        order_proposal.approve()
        order_proposal.apply()
        order_proposal.save()

        order1.refresh_from_db()
        order2.refresh_from_db()
        assert order1.desired_target_weight == Decimal("0.5")
        assert order2.desired_target_weight == Decimal("0.5")
        assert order1.weighting == Decimal("0.5")
        assert order2.weighting == Decimal("0.5")

        order1.desired_target_weight = Decimal("0.7")
        order2.desired_target_weight = Decimal("0.3")
        order1.save()
        order2.save()

        order_proposal.reset_orders(use_desired_target_weight=True)
        order1.refresh_from_db()
        order2.refresh_from_db()
        assert order1.weighting == Decimal("0.7")
        assert order2.weighting == Decimal("0.3")

    def test_reset_order_proposal_keeps_target_cash_weight(self, order_factory, order_proposal_factory):
        order_proposal = order_proposal_factory.create(
            total_cash_weight=Decimal("0.02")
        )  # create a OP with total cash weight of 2%

        # create orders that total weight account for only 50%
        order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.3"))
        order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.2"))

        order_proposal.reset_orders()
        assert order_proposal.get_orders().aggregate(s=Sum("target_weight"))["s"] == Decimal(
            "0.98"
        ), "The total target weight leftover does not equal the stored total cash weight"

    def test_convert_to_portfolio_always_100percent(self, order_proposal, order_factory):
        o1 = order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.5"))
        o2 = order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.3"))

        portfolio = order_proposal.convert_to_portfolio()
        assert portfolio.positions_map[o1.underlying_instrument.id].weighting == Decimal("0.5")
        assert portfolio.positions_map[o2.underlying_instrument.id].weighting == Decimal("0.3")
        assert portfolio.positions_map[order_proposal.cash_component.id].weighting == Decimal("0.2")



import logging
from datetime import datetime

from django.conf import settings
from requests import HTTPError

from wbportfolio.api_clients.ubs import UBSNeoAPIClient
from wbportfolio.pms.typing import Order

from .. import ExecutionStatus, RoutingException
from . import BaseCustodianAdapter

ASSET_CLASS_MAP = {Order.AssetType.EQUITY: "EQUITY"}  # API can support BOND, FUTURE, OPTION, and DYNAMIC_STRATEGY
ASSET_CLASS_MAP_INV = {
    v: k for k, v in ASSET_CLASS_MAP.items()
}  # API can support BOND, FUTURE, OPTION, and DYNAMIC_STRATEGY

STATUS_MAP = {
    "Amend Pending": ExecutionStatus.PENDING,
    "Cancel Pending": ExecutionStatus.PENDING,
    "Cancelled": ExecutionStatus.PENDING,
    "Complete": ExecutionStatus.PENDING,
    "Complete (Order Cancelled)": ExecutionStatus.PENDING,
    "Complete (Partial Fill)": ExecutionStatus.PENDING,
    "In Draft": ExecutionStatus.PENDING,
    "Pending Approval": ExecutionStatus.PENDING,
    "Pending Execution": ExecutionStatus.PENDING,
    "Rebalance Cancelled": ExecutionStatus.PENDING,
    "Rebalance Cancelled (Executing partially)": ExecutionStatus.PENDING,
    "Rejected": ExecutionStatus.PENDING,
    "Rejection Acknowledged": ExecutionStatus.PENDING,
    "Waiting for Response": ExecutionStatus.PENDING,
}
logger = logging.getLogger("oms")


def _serialize_orders(orders: list[Order], default_execution_instruction=None) -> list[dict[str, str]]:
    items = []
    for order in orders:
        if order.refinitiv_identifier_code:
            identifier_type, identifier = "RIC", order.refinitiv_identifier_code
        elif order.bloomberg_ticker:
            identifier_type, identifier = "BBTICKER", order.bloomberg_ticker
        else:
            identifier_type, identifier = "SEDOL", order.sedol
        item = {
            "assetClass": ASSET_CLASS_MAP[order.asset_class],
            "identifierType": identifier_type,
            "identifier": identifier,
            "executionInstruction": order.execution_instruction
            if order.execution_instruction
            else default_execution_instruction,
            "userElementId": str(order.id),
            "tradeDate": order.trade_date.strftime("%Y-%m-%d"),
        }
        if order.shares:
            item["sharesToTrade"] = str(order.shares)
        else:
            item["targetWeight"] = str(order.target_weight)
        items.append(item)
    return items


def _deserialize_items(items: list[dict[str, str]]):
    orders = []
    for item in items:
        orders.append(
            Order(
                id=item.get("userElementId"),
                asset_class=ASSET_CLASS_MAP_INV[item.get("assetClass")],
                refinitiv_identifier_code=item.get(
                    "ric", item["identifier"] if item.get("identifierType") == "RIC" else None
                ),
                bloomberg_ticker=item["identifier"] if item.get("identifierType") == "BBTICKER" else None,
                sedol=item["identifier"] if item.get("identifierType") == "SEDOL" else None,
                trade_date=datetime.strptime(item.get("tradeDate"), "%Y-%m-%d"),
                target_weight=float(item["targetWeight"]) if "targetWeight" in item else None,
                shares=float(item["sharesToTrade"]) if "sharesToTrade" in item else None,
                execution_instruction=item.get("executionInstruction"),
            )
        )
    return orders


class CustodianAdapter(BaseCustodianAdapter):
    client: UBSNeoAPIClient

    def __init__(self, *args, raise_exception: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.raise_exception = raise_exception

    def _handle_response(self, res):
        logger.info(res["message"])
        if errors := res.get("errors"):
            logger.warning(errors)
            if self.raise_exception:
                raise RoutingException(errors)

    def authenticate(self) -> bool:
        """
        Authenticate or renew tokens with the custodian API.
        Raises an exception if authentication fails.
        """
        self.client = UBSNeoAPIClient(settings.UBS_NEO_API_TOKEN)
        return True

    def get_rebalance_status(self) -> tuple[ExecutionStatus, str]:
        res = self.client.get_rebalance_status_for_isin(self.isin)
        self._handle_response(res)
        status = res["rebalanceStatus"]
        return STATUS_MAP.get(status, ExecutionStatus.UNKNOWN), status

    def is_valid(self) -> bool:
        """
        Check whether the given isin is valid and can be rebalanced
        """

        try:
            status_res = self.client.get_rebalance_service_status()

            isin_res = self.client.get_rebalance_status_for_isin(self.isin)
            self._handle_response(status_res)
            self._handle_response(isin_res)
            return (
                status_res["status"] == UBSNeoAPIClient.SUCCESS_VALUE
                and isin_res["status"] == UBSNeoAPIClient.SUCCESS_VALUE
            )
        except (HTTPError, KeyError) as e:
            logger.warning(f"Couldn't validate adapter: {str(e)}")
            return False

    def submit_rebalancing(self, orders: list[Order], as_draft: bool = True) -> tuple[list[Order], str]:
        """
        Submit a rebalance order for the certificate.
        """
        items = _serialize_orders(orders, default_execution_instruction="MARKET_ON_CLOSE")
        if not as_draft:
            res = self.client.submit_rebalance(self.isin, items)
        else:
            res = self.client.save_draft(self.isin, items)
        self._handle_response(res)
        return _deserialize_items(res["rebalanceItems"]), res["message"]

    def cancel_current_rebalancing(self) -> bool:
        """
        Cancel an existing rebalance order identified by ISIN.
        """
        try:
            res = self.client.cancel_rebalance(self.isin)
            self._handle_response(res)
            return res["status"] == UBSNeoAPIClient.SUCCESS_VALUE
        except (HTTPError, KeyError):
            return False

    def get_current_rebalancing(self) -> list[Order]:
        """
        Fetch the current rebalance request details for a certificate.
        """
        res = self.client.get_current_rebalance_request(self.isin)
        self._handle_response(res)
        return _deserialize_items(res["rebalanceItems"])



import logging
import math
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from typing import Any, TypeVar

from celery import shared_task
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import DatabaseError, models
from django.db.models import (
    F,
    OuterRef,
    Subquery,
    Sum,
    Value,
)
from django.db.models.functions import Coalesce, Round
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from pandas._libs.tslibs.offsets import BDay
from wbcompliance.models.risk_management.mixins import RiskCheckMixin
from wbcore.contrib.authentication.models import User
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models import WBModel
from wbcore.utils.models import CloneMixin
from wbfdm.models import InstrumentPrice
from wbfdm.models.instruments.instruments import Cash, Instrument

from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.roles import PortfolioRole
from wbportfolio.pms.trading import TradingService
from wbportfolio.pms.typing import Order as OrderDTO
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position as PositionDTO

from ...order_routing import ExecutionStatus, RoutingException
from ...order_routing.adapters import BaseCustodianAdapter
from .orders import Order
from .routing import cancel_rebalancing, execute_orders, get_execution_status

logger = logging.getLogger("pms")

SelfOrderProposal = TypeVar("SelfOrderProposal", bound="OrderProposal")


class OrderProposal(CloneMixin, RiskCheckMixin, WBModel):
    trade_date = models.DateField(verbose_name="Trading Date")

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        PENDING = "PENDING", "Pending"
        APPROVED = "APPROVED", "Approved"
        DENIED = "DENIED", "Denied"
        APPLIED = "APPLIED", "Applied"
        EXECUTION = "EXECUTION", "Execution"
        FAILED = "FAILED", "Failed"

    comment = models.TextField(default="", verbose_name="Order Comment", blank=True)
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    rebalancing_model = models.ForeignKey(
        "wbportfolio.RebalancingModel",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="order_proposals",
        verbose_name="Rebalancing Model",
        help_text="Rebalancing Model that generates the target portfolio",
    )
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", related_name="order_proposals", on_delete=models.PROTECT, verbose_name="Portfolio"
    )
    creator = models.ForeignKey(
        "directory.Person",
        blank=True,
        null=True,
        related_name="order_proposals",
        on_delete=models.PROTECT,
        verbose_name="Owner",
    )
    approver = models.ForeignKey(
        "directory.Person",
        blank=True,
        null=True,
        related_name="approver_order_proposals",
        on_delete=models.PROTECT,
        verbose_name="Approver",
    )
    min_order_value = models.IntegerField(
        default=0, verbose_name="Minimum Order Value", help_text="Minimum Order Value in the Portfolio currency"
    )
    min_weighting = models.DecimalField(
        max_digits=9,
        decimal_places=Order.ORDER_WEIGHTING_PRECISION,
        default=Decimal(0),
        help_text="The minimum weight allowed for this order proposal ",
        verbose_name="Minimum Weight",
        validators=[MinValueValidator(Decimal("0")), MaxValueValidator(Decimal("1"))],
    )

    total_cash_weight = models.DecimalField(
        default=Decimal("0"),
        decimal_places=4,
        max_digits=5,
        verbose_name="Total Cash Weight",
        help_text="The desired percentage for the cash component. The remaining percentage (100% minus this value) will be allocated to total target weighting. Default is 0%.",
        validators=[MinValueValidator(Decimal("0")), MaxValueValidator(Decimal("1"))],
    )
    total_effective_portfolio_contribution = models.DecimalField(
        default=Decimal("1"),
        max_digits=Order.ORDER_WEIGHTING_PRECISION * 2 + 3,
        decimal_places=Order.ORDER_WEIGHTING_PRECISION * 2,
    )
    execution_status = models.CharField(
        blank=True, default="", choices=ExecutionStatus.choices, verbose_name="Execution Status"
    )
    execution_status_detail = models.CharField(blank=True, default="", verbose_name="Execution Status Detail")
    execution_comment = models.CharField(blank=True, default="", verbose_name="Execution Comment")

    class Meta:
        verbose_name = "Order Proposal"
        verbose_name_plural = "Order Proposals"
        constraints = [
            models.UniqueConstraint(
                fields=["portfolio", "trade_date"],
                name="unique_order_proposal",
            ),
        ]

        notification_types = [
            create_notification_type(
                "wbportfolio.order_proposal.push_model_changes",
                "Push Model Changes",
                "Sends a notification when a the change/orders are pushed to modeled after portfolios",
                True,
                True,
                True,
            )
        ]

    def save(self, *args, **kwargs):
        # if the order proposal is created, we default these fields with the portfolio default value for automatic value assignement
        if not self.id and not self.min_order_value:
            self.min_order_value = self.portfolio.default_order_proposal_min_order_value
        if not self.id and not self.min_weighting:
            self.min_weighting = self.portfolio.default_order_proposal_min_weighting
        if not self.id and not self.total_cash_weight:
            self.total_cash_weight = self.portfolio.default_order_proposal_total_cash_weight
        # if a order proposal is created before the existing earliest order proposal, we automatically shift the linked instruments inception date to allow automatic NAV computation since the new inception date
        if not self.portfolio.order_proposals.filter(trade_date__lt=self.trade_date).exists():
            # we need to set the inception date as the first order proposal trade date (and thus, the first position date). We expect a NAV at 100 then
            self.portfolio.instruments.filter(inception_date__gt=self.trade_date).update(
                inception_date=self.trade_date
            )
        super().save(*args, **kwargs)

    @property
    def check_evaluation_date(self):
        return self.trade_date

    @property
    def checked_object(self) -> Any:
        return self.portfolio

    @cached_property
    def portfolio_total_asset_value(self) -> Decimal:
        return self.portfolio.get_total_asset_value(self.last_effective_date)

    @cached_property
    def validated_trading_service(self) -> TradingService:
        """
        This property holds the validated trading services and cache it.This property expect to be set only if is_valid return True
        """
        target_portfolio = self.convert_to_portfolio()

        return TradingService(
            self.trade_date,
            effective_portfolio=self._get_default_effective_portfolio(),
            target_portfolio=target_portfolio,
            total_target_weight=target_portfolio.total_weight,
        )

    @cached_property
    def last_effective_date(self) -> date:
        try:
            return self.portfolio.assets.filter(date__lt=self.trade_date).latest("date").date
        except AssetPosition.DoesNotExist:
            return self.value_date

    @cached_property
    def custodian_adapter(self) -> BaseCustodianAdapter | None:
        try:
            return self.portfolio.get_authenticated_custodian_adapter(raise_exception=True)
        except ValueError as e:
            logger.warning("Error while instantiating custodian adapter: %s", e)

    @cached_property
    def value_date(self) -> date:
        return (self.trade_date - BDay(1)).date()

    @property
    def previous_order_proposal(self) -> SelfOrderProposal | None:
        future_proposals = OrderProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__lt=self.trade_date, status=OrderProposal.Status.APPLIED
        )
        if future_proposals.exists():
            return future_proposals.latest("trade_date")
        return None

    @property
    def next_order_proposal(self) -> SelfOrderProposal | None:
        future_proposals = OrderProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__gt=self.trade_date, status=OrderProposal.Status.APPLIED
        )
        if future_proposals.exists():
            return future_proposals.earliest("trade_date")
        return None

    @property
    def cash_component(self) -> Cash:
        return self.portfolio.cash_component

    @property
    def total_effective_portfolio_weight(self) -> Decimal:
        return Decimal("1.0")

    @property
    def total_expected_target_weight(self) -> Decimal:
        return self.total_effective_portfolio_weight - self.total_cash_weight

    @cached_property
    def total_effective_portfolio_cash_weight(self) -> Decimal:
        return self.portfolio.assets.filter(
            models.Q(date=self.last_effective_date)
            & (models.Q(underlying_quote__is_cash=True) | models.Q(underlying_quote__is_cash_equivalent=True))
        ).aggregate(Sum("weighting"))["weighting__sum"] or Decimal("0")

    @property
    def execution_instruction(self):
        # TODO make this dynamically configurable
        return "MARKET_ON_CLOSE"

    def get_orders(self):
        # TODO Issue here: the cash is subqueried on the portfolio, on portfolio such as the fund, there is multiple cash component, that we exclude in the orders (and use a unique cash position instead)
        # so the subquery returns the previous position (probably USD), but is missing the other cash aggregation. We need to find a way to handle that properly

        orders = self.orders.all().annotate(
            total_effective_portfolio_contribution=Value(self.total_effective_portfolio_contribution),
            last_effective_date=Subquery(
                AssetPosition.unannotated_objects.filter(
                    date__lt=OuterRef("value_date"),
                    portfolio=OuterRef("portfolio"),
                )
                .order_by("-date")
                .values("date")[:1]
            ),
            previous_weight=models.Case(
                models.When(
                    underlying_instrument__is_cash=False,
                    then=Coalesce(
                        Subquery(
                            AssetPosition.unannotated_objects.filter(
                                underlying_quote=OuterRef("underlying_instrument"),
                                date=OuterRef("last_effective_date"),
                                portfolio=OuterRef("portfolio"),
                            )
                            .values("portfolio")
                            .annotate(s=Sum("weighting"))
                            .values("s")[:1]
                        ),
                        Decimal(0),
                    ),
                ),
                default=Value(self.total_effective_portfolio_cash_weight),
            ),
            contribution=F("previous_weight") * (F("daily_return") + Value(Decimal("1"))),
            effective_weight=Round(
                models.Case(
                    models.When(total_effective_portfolio_contribution=Value(Decimal("0")), then=Value(Decimal("0"))),
                    default=F("contribution") / F("total_effective_portfolio_contribution"),
                ),
                precision=Order.ORDER_WEIGHTING_PRECISION,
            ),
            target_weight=Round(F("effective_weight") + F("weighting"), precision=Order.ORDER_WEIGHTING_PRECISION),
            effective_shares=Coalesce(
                Subquery(
                    AssetPosition.objects.filter(
                        underlying_quote=OuterRef("underlying_instrument"),
                        date=OuterRef("last_effective_date"),
                        portfolio=OuterRef("portfolio"),
                    )
                    .values("portfolio")
                    .annotate(s=Sum("shares"))
                    .values("s")[:1]
                ),
                Decimal(0),
            ),
            target_shares=F("effective_shares") + F("shares"),
        )

        if total_estimated_effective_weight := orders.aggregate(s=models.Sum("effective_weight"))["s"]:
            with suppress(Order.DoesNotExist):
                largest_order = orders.latest("effective_weight")
                if quant_error := self.total_effective_portfolio_weight - total_estimated_effective_weight:
                    orders = orders.annotate(
                        effective_weight=models.Case(
                            models.When(
                                id=largest_order.id,
                                then=models.F("effective_weight") + models.Value(Decimal(quant_error)),
                            ),
                            default=models.F("effective_weight"),
                        ),
                        target_weight=models.Case(
                            models.When(
                                id=largest_order.id,
                                then=models.F("target_weight") + models.Value(Decimal(quant_error)),
                            ),
                            default=models.F("target_weight"),
                        ),
                    )
        return orders.annotate(
            has_warnings=models.Case(
                models.When(
                    (models.Q(price=0) & ~models.Q(target_weight=0)) | models.Q(target_weight__lt=0), then=Value(True)
                ),
                default=Value(False),
            ),
        )

    def prepare_orders_for_execution(self) -> list[OrderDTO]:
        executable_orders = []
        for order in (
            self.get_orders().exclude(underlying_instrument__is_cash=True).select_related("underlying_instrument")
        ):
            instrument = order.underlying_instrument
            # we support only the instrument type provided by the Order DTO class
            asset_class = instrument.get_security_ancestor().instrument_type.key.upper()
            try:
                if instrument.refinitiv_identifier_code or instrument.ticker or instrument.sedol:
                    executable_orders.append(
                        OrderDTO(
                            id=order.id,
                            asset_class=OrderDTO.AssetType[asset_class],
                            weighting=float(order.weighting),
                            target_weight=float(order.target_weight),
                            trade_date=order.value_date,
                            shares=float(order.shares) if order.shares is not None else None,
                            target_shares=float(order.target_shares) if order.target_shares is not None else None,
                            refinitiv_identifier_code=instrument.refinitiv_identifier_code,
                            bloomberg_ticker=instrument.ticker,
                            sedol=instrument.sedol,
                            execution_instruction=self.execution_instruction,
                        )
                    )
                else:
                    order.execution_confirmed = False
                    order.execution_comment = "Underlying instrument does not have a valid identifier."
                    order.save()
            except (AttributeError, KeyError):
                order.execution_confirmed = False
                order.execution_comment = f"Unsupported asset class {asset_class.title()}."
                order.save()

        return executable_orders

    def __str__(self) -> str:
        return f"{self.portfolio.name}: {self.trade_date} ({self.status})"

    def convert_to_portfolio(
        self, use_effective: bool = False, with_cash: bool = True, use_desired_target_weight: bool = False
    ) -> PortfolioDTO:
        """
        Converts the internal portfolio state and pending orders into a PortfolioDTO.

        Args:
            use_effective: Use effective share quantities for positions if True.
            with_cash: Include a cash position in the result if True.
            use_desired_target_weight: Use desired target weights from orders if True.

        Returns:
            PortfolioDTO: Object that encapsulates all portfolio positions.
        """
        portfolio = {}

        # 1. Gather all non-cash, positively weighted assets from the existing portfolio.
        for asset in self.portfolio.assets.filter(
            date=self.last_effective_date,
            # underlying_quote__is_cash=False,
            # underlying_quote__is_cash_equivalent=False,
            weighting__gt=0,
        ):
            portfolio[asset.underlying_quote] = {
                "shares": asset._shares,
                "weighting": asset.weighting,
                "delta_weight": Decimal("0"),
                "price": asset._price,
                "currency_fx_rate": asset._currency_fx_rate,
            }

        # 2. Add or update non-cash orders, possibly overriding weights.
        for order in self.get_orders().filter(
            underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ):
            if use_desired_target_weight and order.desired_target_weight is not None:
                delta_weight = Decimal("0")
                weighting = order.desired_target_weight
            else:
                delta_weight = order.weighting
                weighting = order._previous_weight

            portfolio[order.underlying_instrument] = {
                "weighting": weighting,
                "delta_weight": delta_weight,
                "shares": order._target_shares if not use_effective else order._effective_shares,
                "price": order.price,
                "currency_fx_rate": order.currency_fx_rate,
            }

        # 3. Prepare a mapping from instrument IDs to weights for analytic calculations.
        previous_weights = {instrument.id: float(info["weighting"]) for instrument, info in portfolio.items()}

        # 4. Attempt to fetch analytic returns and portfolio contribution. Default on error.
        try:
            last_returns, contribution = self.portfolio.get_analytic_portfolio(
                self.value_date, weights=previous_weights, use_dl=True
            ).get_contributions()
            last_returns = last_returns.to_dict()
        except ValueError:
            last_returns, contribution = {}, 1

        positions = []
        total_weighting = Decimal("0")

        # 5. Build PositionDTO objects for all instruments.
        for instrument, row in portfolio.items():
            weighting = row["weighting"]
            daily_return = Decimal(last_returns.get(instrument.id, 0))

            # Optionally apply drift to weightings if required
            if not use_effective and not use_desired_target_weight:
                if contribution:
                    drifted_weight = round(
                        weighting * (daily_return + Decimal("1")) / Decimal(contribution),
                        Order.ORDER_WEIGHTING_PRECISION,
                    )
                else:
                    drifted_weight = weighting
                weighting = drifted_weight + row["delta_weight"]

            # Assemble the position object
            positions.append(
                PositionDTO(
                    underlying_instrument=instrument.id,
                    instrument_type=instrument.instrument_type.id,
                    weighting=weighting,
                    daily_return=daily_return if use_effective else Decimal("0"),
                    shares=row["shares"],
                    currency=instrument.currency.id,
                    date=self.last_effective_date if use_effective else self.trade_date,
                    is_cash=instrument.is_cash or instrument.is_cash_equivalent,
                    price=row["price"],
                    currency_fx_rate=row["currency_fx_rate"],
                )
            )
            total_weighting += weighting

        # 6. Optionally include a cash position to balance the total weighting.
        if (
            portfolio
            and with_cash
            and total_weighting
            and self.total_effective_portfolio_weight
            and (cash_weight := self.total_effective_portfolio_weight - total_weighting)
        ):
            cash_position = self.get_estimated_target_cash(target_cash_weight=cash_weight)
            positions.append(cash_position._build_dto())

        return PortfolioDTO(positions)

    # Start tools methods
    def _clone(self, **kwargs) -> SelfOrderProposal:
        """
        Method to clone self as a new order proposal. It will automatically shift the order date if a proposal already exists
        Args:
            **kwargs: The keyword arguments
        Returns:
            The cloned order proposal
        """
        trade_date = kwargs.get("clone_date", self.trade_date)

        # Find the next valid order date
        while OrderProposal.objects.filter(portfolio=self.portfolio, trade_date=trade_date).exists():
            trade_date += timedelta(days=1)

        order_proposal_clone = OrderProposal.objects.create(
            trade_date=trade_date,
            comment=kwargs.get("clone_comment", self.comment),
            status=OrderProposal.Status.DRAFT,
            rebalancing_model=self.rebalancing_model,
            portfolio=self.portfolio,
            creator=self.creator,
        )
        for order in self.orders.all():
            order.id = None
            order.order_proposal = order_proposal_clone
            order.save()

        return order_proposal_clone

    def normalize_orders(self):
        """
        Call the trading service with the existing orders and normalize them in order to obtain a total sum target weight of 100%
        The existing order will be modified directly with the given normalization factor
        """
        service = TradingService(
            self.trade_date,
            effective_portfolio=self._get_default_effective_portfolio(),
            target_portfolio=self.convert_to_portfolio(use_effective=False, with_cash=False),
            total_target_weight=self.total_expected_target_weight,
        )
        leftovers_orders = self.orders.all()
        portfolio_value = self.portfolio_total_asset_value
        for underlying_instrument_id, order_dto in service.trades_batch.trades_map.items():
            with suppress(Order.DoesNotExist):
                order = self.orders.get(underlying_instrument_id=underlying_instrument_id)
                order.set_weighting(round(order_dto.delta_weight, Order.ORDER_WEIGHTING_PRECISION), portfolio_value)
                order.save()
                leftovers_orders = leftovers_orders.exclude(id=order.id)
        leftovers_orders.delete()
        self.fix_quantization()

    def fix_quantization(self):
        if self.orders.exists():
            orders = self.get_orders()
            portfolio_value = self.portfolio_total_asset_value
            t_weight = orders.aggregate(models.Sum("target_weight"))["target_weight__sum"] or Decimal("0.0")
            # we handle quantization error due to the decimal max digits. In that case, we take the biggest order (highest weight) and we remove the quantization error
            if quantize_error := (t_weight - self.total_expected_target_weight):
                biggest_order = orders.exclude(underlying_instrument__is_cash=True).latest("target_weight")
                biggest_order.set_weighting(biggest_order.weighting - quantize_error, portfolio_value)
                biggest_order.save()

    def _get_default_target_portfolio(self, use_desired_target_weight: bool = False, **kwargs) -> PortfolioDTO:
        if self.rebalancing_model:
            params = {}
            if rebalancer := getattr(self.portfolio, "automatic_rebalancer", None):
                params.update(rebalancer.parameters)
            params.update(kwargs)
            return self.rebalancing_model.get_target_portfolio(
                self.portfolio, self.trade_date, self.value_date, **params
            )
        return self.convert_to_portfolio(use_effective=False, use_desired_target_weight=use_desired_target_weight)

    def _get_default_effective_portfolio(self):
        return self.convert_to_portfolio(use_effective=True)

    def reset_orders(
        self,
        target_portfolio: PortfolioDTO | None = None,
        effective_portfolio: PortfolioRole | None = None,
        validate_order: bool = True,
        use_desired_target_weight: bool = False,
    ):
        """
        Will delete all existing orders and recreate them from the method `create_or_update_trades`
        """
        if self.rebalancing_model:
            self.orders.all().delete()
        else:
            self.orders.filter(underlying_instrument__is_cash=True).delete()
        # delete all existing orders
        # Get effective and target portfolio
        if not target_portfolio:
            target_portfolio = self._get_default_target_portfolio(use_desired_target_weight=use_desired_target_weight)
        if not effective_portfolio:
            effective_portfolio = self._get_default_effective_portfolio()
        if target_portfolio:
            service = TradingService(
                self.trade_date,
                effective_portfolio=effective_portfolio,
                target_portfolio=target_portfolio,
                total_target_weight=self.total_expected_target_weight,
            )
            if validate_order:
                service.is_valid()
                orders = service.validated_trades
            else:
                orders = service.trades_batch.trades_map.values()

            objs = []
            portfolio_value = self.portfolio_total_asset_value
            for order_dto in orders:
                instrument = Instrument.objects.get(id=order_dto.underlying_instrument)
                currency_fx_rate = instrument.currency.convert(
                    self.value_date, self.portfolio.currency, exact_lookup=True
                )
                # we cannot do a bulk-create because Order is a multi table inheritance
                weighting = round(order_dto.delta_weight, Order.ORDER_WEIGHTING_PRECISION)
                daily_return = order_dto.daily_return
                try:
                    order = self.orders.get(underlying_instrument=instrument)
                    order.currency_fx_rate = currency_fx_rate
                    order.daily_return = daily_return
                except Order.DoesNotExist:
                    order = Order(
                        underlying_instrument=instrument,
                        order_proposal=self,
                        value_date=self.trade_date,
                        weighting=weighting,
                        daily_return=daily_return,
                        currency_fx_rate=currency_fx_rate,
                    )
                order.pre_save()
                order.set_weighting(weighting, portfolio_value)
                order.desired_target_weight = order_dto.target_weight
                order.order_type = Order.get_type(
                    weighting, round(order_dto.previous_weight, 8), round(order_dto.target_weight, 8)
                )
                # # if we cannot automatically find a price, we consider the stock is invalid and we sell it
                # if not order.price and order.weighting > 0:
                #     order.price = Decimal("0.0")
                #     order.weighting = -order_dto.effective_weight
                objs.append(order)
            Order.objects.bulk_create(
                objs,
                update_fields=[
                    "value_date",
                    "weighting",
                    "daily_return",
                    "currency_fx_rate",
                    "order_type",
                    "portfolio",
                    "price",
                    "price_gross",
                    "desired_target_weight",
                    "shares",
                ],
                unique_fields=["order_proposal", "underlying_instrument"],
                update_conflicts=True,
                batch_size=1000,
            )
        # final sanity check to make sure invalid order with effective and target weight of 0 are automatically removed:
        self.get_orders().exclude(underlying_instrument__is_cash=True).filter(
            target_weight=0, effective_weight=0
        ).delete()
        self.get_orders().filter(target_weight=0).exclude(effective_shares=0).update(shares=-F("effective_shares"))
        # self.fix_quantization()
        self.total_effective_portfolio_contribution = effective_portfolio.portfolio_contribution
        self.save()

    def apply_workflow(
        self,
        apply_automatically: bool = True,
        silent_exception: bool = False,
        force_reset_order: bool = False,
        **reset_order_kwargs,
    ):
        # before, we need to save all positions in the builder first because effective weight depends on it
        self.portfolio.builder.bulk_create_positions(delete_leftovers=True)
        if self.status == OrderProposal.Status.APPLIED:
            logger.info("Reverting order proposal ...")
            self.revert()
        if self.status == OrderProposal.Status.DRAFT:
            if (
                self.rebalancing_model or force_reset_order
            ):  # if there is no position (for any reason) or we the order proposal has a rebalancer model attached (orders are computed based on an aglo), we reapply this order proposal
                logger.info("Resetting orders ...")
                try:  # we silent any validation error while setting proposal, because if this happens, we assume the current order proposal state if valid and we continue to batch compute
                    self.reset_orders(**reset_order_kwargs)
                except (ValidationError, DatabaseError) as e:
                    self.status = OrderProposal.Status.FAILED
                    if not silent_exception:
                        raise ValidationError(e) from e
                    return
            logger.info("Submitting order proposal ...")
            self.submit()
        if self.status == OrderProposal.Status.PENDING:
            self.approve()
        if apply_automatically and self.portfolio.can_be_rebalanced:
            logger.info("Applying order proposal ...")
            self.apply(replay=False)

    def replay(
        self,
        broadcast_changes_at_date: bool = True,
        reapply_order_proposal: bool = False,
        synchronous: bool = False,
        **reset_order_kwargs,
    ):
        last_order_proposal = self
        last_order_proposal_created = False
        self.portfolio.load_builder_returns((self.trade_date - BDay(3)).date(), date.today())
        while last_order_proposal and last_order_proposal.status == OrderProposal.Status.APPLIED:
            last_order_proposal.portfolio = self.portfolio  # we set the same ptf reference
            if not last_order_proposal_created:
                if reapply_order_proposal or last_order_proposal.rebalancing_model:
                    logger.info(f"Replaying order proposal {last_order_proposal}")
                    last_order_proposal.apply_workflow(
                        silent_exception=True, force_reset_order=True, **reset_order_kwargs
                    )
                    last_order_proposal.save()
                else:
                    logger.info(f"Resetting order proposal {last_order_proposal}")
                    last_order_proposal.reset_orders(**reset_order_kwargs)
                if last_order_proposal.status != OrderProposal.Status.APPLIED:
                    break
            next_order_proposal = last_order_proposal.next_order_proposal
            if next_order_proposal:
                next_trade_date = next_order_proposal.trade_date - timedelta(days=1)
            elif next_expected_rebalancing_date := self.portfolio.get_next_rebalancing_date(
                last_order_proposal.trade_date
            ):
                next_trade_date = (
                    next_expected_rebalancing_date + timedelta(days=7)
                )  # we don't know yet if rebalancing is valid and can be executed on `next_expected_rebalancing_date`, so we add safety window of 7 days
            else:
                next_trade_date = date.today()
            next_trade_date = min(next_trade_date, date.today())
            gen = self.portfolio.drift_weights(
                last_order_proposal.trade_date, next_trade_date, stop_at_rebalancing=True
            )
            try:
                while True:
                    self.portfolio.builder.add(next(gen))
            except StopIteration as e:
                overriding_order_proposal = e.value

            self.portfolio.builder.bulk_create_positions(
                delete_leftovers=True,
            )
            for draft_tp in OrderProposal.objects.filter(
                portfolio=self.portfolio,
                trade_date__gt=last_order_proposal.trade_date,
                trade_date__lte=next_trade_date,
                status=OrderProposal.Status.DRAFT,
            ):
                draft_tp.reset_orders()
            if overriding_order_proposal:
                last_order_proposal_created = True
                last_order_proposal = overriding_order_proposal
            else:
                last_order_proposal_created = False
                last_order_proposal = next_order_proposal
        self.portfolio.builder.schedule_change_at_dates(
            synchronous=synchronous, broadcast_changes_at_date=broadcast_changes_at_date
        )

    def invalidate_future_order_proposal(self):
        # Delete all future automatic order proposals and set the manual one into a draft state
        self.portfolio.order_proposals.filter(
            trade_date__gt=self.trade_date, rebalancing_model__isnull=False, comment="Automatic rebalancing"
        ).delete()
        for future_order_proposal in self.portfolio.order_proposals.filter(
            trade_date__gt=self.trade_date, status=OrderProposal.Status.APPLIED
        ):
            future_order_proposal.revert()
            future_order_proposal.save()

    def get_estimated_shares(
        self, weight: Decimal, underlying_quote: Instrument, quote_price: Decimal
    ) -> Decimal | None:
        """
        Estimates the number of shares for a order based on the given weight and underlying quote.

        This method calculates the estimated shares by dividing the order's total value in the portfolio's currency by the price of the underlying quote in the same currency. It handles currency conversion and suppresses any ValueError that might occur during the price retrieval.

        Args:
            weight (Decimal): The weight of the order.
            underlying_quote (Instrument): The underlying instrument for the order.

        Returns:
            Decimal | None: The estimated number of shares or None if the calculation fails.
        """
        # Retrieve the price of the underlying quote on the order date TODO: this is very slow and probably due to the to_date argument to the dl which slowdown drastically the query

        # if an order exists for this estimation and the target weight is 0, then we return the inverse of the effective shares
        with suppress(Order.DoesNotExist):
            order = self.get_orders().get(underlying_instrument=underlying_quote)
            if order.target_weight == 0:
                return -order.effective_shares
        # Calculate the order's total value in the portfolio's currency
        trade_total_value_fx_portfolio = self.portfolio_total_asset_value * weight

        # Convert the quote price to the portfolio's currency
        price_fx_portfolio = quote_price * underlying_quote.currency.convert(
            self.trade_date, self.portfolio.currency, exact_lookup=False
        )
        # If the price is valid, calculate and return the estimated shares
        if price_fx_portfolio:
            return trade_total_value_fx_portfolio / price_fx_portfolio

    def get_round_lot_size(self, shares: Decimal, underlying_quote: Instrument) -> Decimal:
        if (round_lot_size := underlying_quote.round_lot_size) != 1 and (
            not underlying_quote.exchange or underlying_quote.exchange.apply_round_lot_size
        ):
            if shares > 0:
                shares = math.ceil(shares / round_lot_size) * round_lot_size
            elif abs(shares) > round_lot_size:
                shares = math.floor(shares / round_lot_size) * round_lot_size
        return shares

    def get_estimated_target_cash(self, target_cash_weight: Decimal | None = None) -> AssetPosition:
        """
        Estimates the target cash weight and shares for a order proposal.

        This method calculates the target cash weight by summing the weights of cash orders and adding any leftover weight from non-cash orders. It then estimates the target shares for this cash component if the portfolio is not only weighting-based.

        Args:
            target_cash_weight (Decimal): the expected target cash weight (Optional). If not provided, we estimate from the existing orders

        Returns:
            tuple[Decimal, Decimal]: A tuple containing the target cash weight and the estimated target shares.
        """
        # Retrieve orders with base information
        orders = self.get_orders()
        # Calculate the total target weight of all orders
        total_target_weight = orders.filter(
            underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ).aggregate(s=models.Sum("target_weight"))["s"] or Decimal(0)
        if target_cash_weight is None:
            target_cash_weight = Decimal("1") - total_target_weight

        # Initialize target shares to zero
        total_target_shares = Decimal(0)

        # Get or create a cash component for the portfolio's currency
        cash_component = self.cash_component
        # If the portfolio is not only weighting-based, estimate the target shares for the cash component
        if not self.portfolio.only_weighting:
            # Estimate the target shares for the cash component
            with suppress(ValueError):
                total_target_shares = self.get_estimated_shares(target_cash_weight, cash_component, Decimal("1.0"))

        # otherwise, we create a new position
        underlying_quote_price = InstrumentPrice.objects.get_or_create(
            instrument=cash_component,
            date=self.trade_date,
            calculated=False,
            defaults={"net_value": Decimal(1)},
        )[0]
        return AssetPosition(
            underlying_quote=cash_component,
            portfolio_created=None,
            portfolio=self.portfolio,
            date=self.trade_date,
            weighting=target_cash_weight,
            initial_price=underlying_quote_price.net_value,
            initial_shares=total_target_shares,
            asset_valuation_date=self.trade_date,
            underlying_quote_price=underlying_quote_price,
            currency=cash_component.currency,
            is_estimated=False,
        )

    # Start FSM logics

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.PENDING,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submit",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        orders = []
        orders_validation_warnings = []
        qs = self.get_orders()
        total_target_weight = Decimal("0")
        for order in qs:
            order_warnings = order.submit(
                by=by, description=description, portfolio_total_asset_value=self.portfolio_total_asset_value, **kwargs
            )

            if order_warnings:
                orders_validation_warnings.extend(order_warnings)
            orders.append(order)
            total_target_weight += order._target_weight

        Order.objects.bulk_update(orders, ["shares", "weighting", "desired_target_weight"])

        # If we estimate cash on this order proposal, we make sure to create the corresponding cash component
        estimated_cash_position = self.get_estimated_target_cash()
        target_portfolio = self.validated_trading_service.trades_batch.convert_to_portfolio(
            estimated_cash_position._build_dto()
        )
        self.evaluate_active_rules(self.trade_date, target_portfolio, asynchronously=True)
        self.total_cash_weight = Decimal("1") - total_target_weight
        return orders_validation_warnings

    def can_submit(self):
        errors = dict()
        errors_list = []
        service = self.validated_trading_service
        try:
            service.is_valid(ignore_error=True)
            # if service.trades_batch.total_abs_delta_weight == 0:
            #     errors_list.append(
            #         "There is no change detected in this order proposal. Please submit at last one valid order"
            #     )
            if len(service.validated_trades) == 0:
                errors_list.append(_("There is no valid order on this proposal"))
            if service.errors:
                errors_list.extend(service.errors)
            if errors_list:
                errors["non_field_errors"] = errors_list
        except ValidationError:
            errors["non_field_errors"] = service.errors
            with suppress(KeyError):
                del self.__dict__["validated_trading_service"]
        return errors

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.APPROVED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and not instance.has_non_successful_checks,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approve",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def approve(self, by=None, description=None, **kwargs):
        if by:
            self.approver = getattr(by, "profile", None)
        elif not self.approver:
            self.approver = self.creator

    def can_approve(self):
        pass

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.DENIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and not instance.has_non_successful_checks,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        pass

    def can_deny(self):
        pass

    @property
    def can_be_applied(self):
        return not self.has_non_successful_checks and self.portfolio.is_manageable

    @transition(
        field=status,
        source=Status.APPROVED,
        target=Status.APPLIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.can_be_applied,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.SAVE.icon,
                key="apply",
                label="Apply",
                action_label="Apply",
            )
        },
    )
    def apply(self, by=None, description=None, replay: bool = True, **kwargs):
        # We validate order which will create or update the initial asset positions
        if not self.portfolio.can_be_rebalanced:
            raise ValueError("Non-Rebalanceable portfolio cannot be traded manually.")
        warnings = []
        # We do not want to create the estimated cash position if there is not orders in the order proposal (shouldn't be possible anyway)
        estimated_cash_position = self.get_estimated_target_cash()
        assets = {}
        for order in self.get_orders():
            with suppress(ValueError):
                # we add the corresponding asset only if it is not the cache position (already included in estimated_cash_position)
                if (
                    order.underlying_instrument != estimated_cash_position.underlying_quote
                    and order._target_weight > 0
                ):
                    assets[order.underlying_instrument.id] = order._target_weight

        # if there is cash leftover, we create an extra asset position to hold the cash component
        if estimated_cash_position.weighting > 0 and len(assets) > 0:
            warnings.append(
                f"We created automatically a cash position of weight {estimated_cash_position.weighting:.2%}"
            )
            assets[estimated_cash_position.underlying_quote.id] = estimated_cash_position.weighting
        self.portfolio.builder.add((self.trade_date, assets)).bulk_create_positions(
            force_save=True, is_estimated=False, delete_leftovers=True
        )
        if replay and self.portfolio.is_manageable:
            replay_as_task.delay(
                self.id, user_id=by.id if by else None, broadcast_changes_at_date=False, reapply_order_proposal=True
            )
        if by:
            self.approver = by.profile
        return warnings

    def can_apply(self):
        errors = dict()
        orders = self.get_orders()
        if not self.portfolio.can_be_rebalanced:
            errors["non_field_errors"] = [_("The portfolio does not allow manual rebalanced")]
        if not orders.exists():
            errors["non_field_errors"] = [
                _("At least one order needs to be submitted to be able to apply this proposal")
            ]
        if not self.portfolio.can_be_rebalanced:
            errors["portfolio"] = [
                [_("The portfolio needs to be a model portfolio in order to apply this order proposal manually")]
            ]
        if self.has_non_successful_checks:
            errors["non_field_errors"] = [_("The pre orders rules did not passed successfully")]
        if orders.filter(has_warnings=True).filter(
            underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ):
            errors["non_field_errors"] = [
                _("There is warning that needs to be addresses on the orders before approval.")
            ]
        return errors

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.has_all_check_completed
        or not instance.checks.exists(),  # we wait for all checks to succeed before proposing the back to draft transition
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="backtodraft",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        self.checks.delete()

    def can_backtodraft(self):
        pass

    @transition(
        field=status,
        source=Status.APPLIED,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.REGENERATE.icon,
                key="revert",
                label="Revert",
                action_label="revert",
                description_fields="<p>Unapply orders and move everything back to draft (i.e. The underlying asset positions will change like the orders were never applied)</p>",
            )
        },
    )
    def revert(self, **kwargs):
        self.approver = None
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        self.portfolio.assets.filter(date=self.trade_date, is_estimated=False).update(
            is_estimated=True
        )  # we delete the existing portfolio as it has been reverted

    def can_revert(self):
        errors = dict()
        if not self.portfolio.can_be_rebalanced:
            errors["portfolio"] = [
                _("The portfolio needs to be a model portfolio in order to revert this order proposal manually")
            ]
        return errors

    @transition(
        field=status,
        source=Status.APPROVED,
        target=Status.EXECUTION,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and (user != instance.approver or user.is_superuser)
        and instance.custodian_adapter
        and not instance.has_non_successful_checks,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.DEAL_MONEY.icon,
                key="execute",
                label="Execute",
                action_label="Execute",
                description_fields="<p>Execute the orders through the setup custodian.</p>",
            )
        },
    )
    def execute(self, **kwargs):
        self.execution_status = ExecutionStatus.PENDING
        self.execution_comment = "Waiting for custodian confirmation"
        execute_orders_as_task.delay(self.id)

    def can_execute(self):
        if not self.custodian_adapter:
            return {"portfolio": ["No custodian adapter"]}

    @transition(
        field=status,
        source=Status.EXECUTION,
        target=Status.APPROVED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.execution_status in [ExecutionStatus.PENDING, ExecutionStatus.IN_DRAFT],
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:orderproposal",),
                icon=WBIcon.PREVIOUS.icon,
                key="cancelexecution",
                label="Cancel Execution",
                action_label="Cancel Execution",
                description_fields="<p>Cancel the current requested execution. Time sensitive operation.</p>",
            )
        },
    )
    def cancelexecution(self, **kwargs):
        warning = ""
        try:
            if cancel_rebalancing(self):
                self.execution_comment, self.execution_status_detail, self.execution_status = (
                    "",
                    "",
                    ExecutionStatus.CANCELLED,
                )
            else:
                warning = "We could not cancel the rebalancing. It is probably already executed. Please refresh status or check with an administrator."
        except (RoutingException, ValueError) as e:
            warning = f"Could not cancel orders proposal {self}: {str(e)}"
            logger.error(warning)
        return warning

    def can_cancelexecution(self):
        if self.execution_status not in [ExecutionStatus.PENDING, ExecutionStatus.IN_DRAFT]:
            return {"execution_status": "Execution can only be cancelled if it is not already executed"}

    def update_execution_status(self):
        try:
            self.execution_status, self.execution_status_detail = get_execution_status(self)
            self.save()
        except (RoutingException, ValueError) as e:
            logger.warning(f"Could not update rebalancing status: {str(e)}")

    # End FSM logics

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbportfolio:orderproposal"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:orderproposalrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{_portfolio.name}} ({{trade_date}})"


@receiver(post_save, sender="wbportfolio.OrderProposal")
def post_fail_order_proposal(sender, instance: OrderProposal, created, raw, **kwargs):
    # if we have a order proposal in a fail state, we ensure that all future existing order proposal are either deleted (automatic one) or set back to draft
    if not raw and instance.status == OrderProposal.Status.FAILED:
        # we delete all order proposal that have a rebalancing model and are marked as "automatic" (quite hardcoded yet)
        instance.invalidate_future_order_proposal()


@shared_task(queue="portfolio")
def replay_as_task(order_proposal_id, user_id: int | None = None, **kwargs):
    order_proposal = OrderProposal.objects.get(id=order_proposal_id)
    order_proposal.replay(**kwargs)
    if user_id:
        user = User.objects.get(id=user_id)
        send_notification(
            code="wbportfolio.portfolio.replay_done",
            title="Order Proposal Replay Completed",
            body=f'We’ve successfully replayed your order proposal for "{order_proposal.portfolio}" from {order_proposal.trade_date:%Y-%m-%d}. You can now review its updated composition.',
            user=user,
            reverse_name="wbportfolio:portfolio-detail",
            reverse_args=[order_proposal.portfolio.id],
        )


@shared_task(queue="portfolio")
def execute_orders_as_task(order_proposal_id: int):
    order_proposal = OrderProposal.objects.get(id=order_proposal_id)
    try:
        status, comment = execute_orders(order_proposal)
    except (ValueError, RoutingException) as e:
        logger.error(f"Could not execute orders proposal {order_proposal}: {str(e)}")
        status = ExecutionStatus.FAILED
        comment = str(e)
    order_proposal.execution_status = status
    order_proposal.execution_comment = comment
    order_proposal.save()


@shared_task(queue="portfolio")
def push_model_change_as_task(
    model_order_proposal_id: int,
    user_id: int,
    only_for_portfolio_ids: list[int] | None = None,
    approve_automatically: bool = False,
):
    # not happy with that but we will keep it for the MVP lifecycle
    model_order_proposal = OrderProposal.objects.get(id=model_order_proposal_id)
    trade_date = model_order_proposal.trade_date
    user = User.objects.get(id=user_id)
    from wbportfolio.models.rebalancing import RebalancingModel

    model_rebalancing = RebalancingModel.objects.get(
        class_path="wbportfolio.rebalancing.models.model_portfolio.ModelPortfolioRebalancing"
    )
    product_html_list = "<ul>\n"
    for rel in model_order_proposal.portfolio.get_model_portfolio_relationships(trade_date):
        if not only_for_portfolio_ids or rel.portfolio.id in only_for_portfolio_ids:
            order_proposal, _ = OrderProposal.objects.update_or_create(
                portfolio=rel.portfolio, trade_date=trade_date, defaults={"rebalancing_model": model_rebalancing}
            )
            order_proposal.reset_orders()
            product_html_list += f"<li>{rel.portfolio}</li>\n"
            if approve_automatically:
                order_proposal.submit()
                order_proposal.approve(by=user)
                order_proposal.save()
    product_html_list += "</ul>"

    send_notification(
        code="wbportfolio.order_proposal.push_model_changes",
        title="Portfolio Model changes are pushed to dependant portfolios",
        body=f"""
<p>The latest updates to the portfolio model <strong>{model_order_proposal.portfolio}</strong> have been successfully applied to the associated portfolios, and corresponding orders have been created.</p>
<p>To proceed with executing these orders, please review the following related portfolios: </p>
        {product_html_list}
        """,
        user=user,
    )



import logging
import math
from collections import defaultdict
from datetime import date
from decimal import Decimal
from typing import TYPE_CHECKING, Iterable

import pandas as pd
from celery import chain, group
from django.contrib.contenttypes.models import ContentType
from pandas._libs.tslibs.offsets import BDay
from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbfdm.contrib.metric.tasks import compute_metrics_as_task
from wbfdm.models import Instrument

from wbportfolio.models.asset import AssetPosition
from wbportfolio.pms.analytics.portfolio import Portfolio as AnalyticPortfolio

if TYPE_CHECKING:
    from wbportfolio.models import Portfolio
logger = logging.getLogger("pms")


MINIMUM_DECIMAL = 8
MIN_STEP = Decimal("0.00000001")


class AssetPositionBuilder:
    """
    Efficiently converts position data into AssetPosition models with batch operations
    and proper dependency management.

    Features:
    - Bulk database fetching for performance
    - Thread-safe operations
    - Clear type hints
    - Memory-efficient storage
    """

    _positions: dict[date, dict[tuple[int, int | None], "AssetPosition"]]

    _fx_rates: dict[date, dict[Currency, CurrencyFXRates]]
    _instruments: dict[int, Instrument]

    def __init__(
        self,
        portfolio: "Portfolio",
    ):
        self.portfolio = portfolio
        # Initialize data stores with type hints
        self._instruments = {}
        self._fx_rates = defaultdict(dict)
        self.prices = defaultdict(dict)
        self.returns = pd.DataFrame()
        self._compute_metrics_tasks = set()
        self._change_at_date_tasks = dict()
        self._positions = defaultdict(dict)

    def get_positions(self, fix_quantization: bool = True, **kwargs):
        # return an iterable excluding the position with a null weight if the portfolio is manageable (otherwise, we assume the 0-weight position is valid)
        for positions in self._positions.values():
            quantization_weight_error = round(
                Decimal("1") - sum(map(lambda o: o.weighting, positions.values()))
                if fix_quantization
                else Decimal("0"),
                MINIMUM_DECIMAL,
            )
            for position in sorted(positions.values(), key=lambda x: x.weighting, reverse=True):
                if position.weighting:
                    for k, v in kwargs.items():
                        setattr(position, k, v)
                    # if the total weight is not 100%, we add the quantization leftover to some random position (max 1e-8 per position, thus it is negligible)
                    if quantization_weight_error:
                        step = round(Decimal(math.copysign(MIN_STEP, quantization_weight_error)), MINIMUM_DECIMAL)
                        position.weighting += step
                        quantization_weight_error -= step
                    yield position

    def __bool__(self) -> bool:
        return len(self._positions.keys()) > 0

    def _get_instrument(self, instrument_id: int) -> Instrument:
        try:
            return self._instruments[instrument_id]
        except KeyError:
            instrument = Instrument.objects.get(id=instrument_id)
            self._instruments[instrument_id] = instrument
            return instrument

    def _get_fx_rate(self, val_date: date, currency: Currency) -> CurrencyFXRates | None:
        try:
            return self._fx_rates[val_date][currency]
        except KeyError:
            if currency.key == "USD":
                fx_rate = CurrencyFXRates.objects.get_or_create(
                    currency=currency, date=val_date, defaults={"value": Decimal("1")}
                )[0]
            else:
                try:
                    fx_rate = CurrencyFXRates.objects.get(
                        currency=currency, date=val_date
                    )  # we create a fx rate anyway to not fail the position. The fx rate expect to be there later on
                except CurrencyFXRates.DoesNotExist:
                    fx_rate = CurrencyFXRates.objects.filter(currency=currency, date__lt=val_date).latest("date")
            self._fx_rates[val_date][currency] = fx_rate
            return fx_rate

    def _get_price(self, val_date: date, instrument: Instrument) -> float | None:
        try:
            return self.prices[val_date][instrument.id]
        except KeyError:
            return None

    def _dict_to_model(
        self,
        val_date: date,
        instrument_id: int,
        weighting: float,
        **kwargs,
    ) -> "AssetPosition":
        underlying_quote = self._get_instrument(instrument_id)
        currency_fx_rate_portfolio_to_usd = self._get_fx_rate(val_date, self.portfolio.currency)
        currency_fx_rate_instrument_to_usd = self._get_fx_rate(val_date, underlying_quote.currency)
        price = self._get_price(val_date, underlying_quote)

        parameters = dict(
            underlying_quote=underlying_quote,
            weighting=round(weighting, MINIMUM_DECIMAL),
            date=val_date,
            asset_valuation_date=val_date,
            is_estimated=True,
            portfolio=self.portfolio,
            currency=underlying_quote.currency,
            initial_price=price,
            currency_fx_rate_portfolio_to_usd=currency_fx_rate_portfolio_to_usd,
            currency_fx_rate_instrument_to_usd=currency_fx_rate_instrument_to_usd,
            initial_currency_fx_rate=None,
            underlying_quote_price=None,
            underlying_instrument=None,
        )
        parameters.update(kwargs)
        position = AssetPosition(**parameters)
        return position

    def load_returns(self, instrument_ids: Iterable[int], from_date: date, to_date: date, use_dl: bool = True):
        if self.returns.empty:
            self.prices, self.returns = Instrument.objects.filter(id__in=instrument_ids).get_returns_df(
                from_date=from_date, to_date=to_date, to_currency=self.portfolio.currency, use_dl=use_dl
            )
        else:
            min_date = min(self.prices.keys())
            max_date = max(self.prices.keys())
            if from_date < min_date or to_date > max_date:
                # we need to refetch everything as we are missing index
                self.prices, self.returns = Instrument.objects.filter(
                    id__in=set(instrument_ids).union(set(self.returns.columns))
                ).get_returns_df(
                    from_date=min(from_date, min_date),
                    to_date=max(to_date, max_date),
                    to_currency=self.portfolio.currency,
                    use_dl=use_dl,
                )
            else:
                instruments = set(instrument_ids) - set(self.returns.columns)
                if instruments:
                    new_prices, new_returns = Instrument.objects.filter(id__in=instruments).get_returns_df(
                        from_date=min(from_date, min_date),
                        to_date=max(to_date, max_date),
                        to_currency=self.portfolio.currency,
                        use_dl=use_dl,
                    )
                    self.returns = self.returns.join(new_returns, how="left").fillna(0)
                    for d, p in new_prices.items():
                        self.prices[d].update(p)

    def add(
        self,
        positions: list["AssetPosition"] | tuple[date, dict[int, float]],
        infer_underlying_quote_price: bool = False,
    ):
        """
        Add multiple positions efficiently with batch processing

        Args:
            positions: Iterable of AssetPosition instances or dictionary of weight {instrument_id: weight} that needs to be converted into AssetPosition
        """
        if isinstance(positions, tuple):
            val_date = positions[0]
            positions = [(val_date, i, w) for i, w in positions[1].items()]  # unflatten data to make it iterable
        for position in positions:
            if not isinstance(position, AssetPosition):
                position = self._dict_to_model(*position)
            position.pre_save(
                infer_underlying_quote_price=infer_underlying_quote_price
            )  # inferring underlying quote price is potentially very slow for big dataset of positions, it's not very needed for model portfolio so we disable it
            # Generate unique composite key
            key = (
                position.underlying_quote.id,
                position.portfolio_created.id if position.portfolio_created else None,
            )
            # Merge duplicate positions
            if existing_position := self._positions[position.date].get(key):
                position.weighting += existing_position.weighting
                if existing_position.initial_shares:
                    position.initial_shares += existing_position.initial_shares
            # ensure the position portfolio is the iterator portfolio (could be different when computing look-through for instance)
            position.portfolio = self.portfolio
            position.weighting = Decimal(
                round(position.weighting, 8)
            )  # set the weight as it will be saved in the db to handle quantization error accordingly
            if position.initial_price is not None and position.initial_currency_fx_rate is not None:
                self._positions[position.date][key] = position
        return self

    def get_dates(self) -> list[date]:
        """Get sorted list of unique dates"""
        return list(sorted(self._positions.keys()))

    def _get_portfolio(self, val_date: date) -> AnalyticPortfolio:
        """Get weight structure with instrument IDs as keys"""
        positions = self._positions[val_date]
        next_returns = self.returns.loc[[(val_date + BDay(1)).date()], :]
        weights = dict(map(lambda row: (row[1].underlying_quote.id, float(row[1].weighting)), positions.items()))
        return AnalyticPortfolio(weights=weights, X=next_returns)

    def bulk_create_positions(self, delete_leftovers: bool = False, force_save: bool = False, **kwargs):
        positions = list(self.get_positions(**kwargs))
        # we need to delete the existing estimated portfolio because otherwise we risk to have existing and not
        # overlapping positions remaining (as they will not be updating by the bulk create). E.g. when someone
        # change completely the trades of a portfolio model and drift it.
        dates = self.get_dates()
        self.portfolio.assets.filter(date__in=dates, is_estimated=True).delete()

        if len(positions) > 0:
            if self.portfolio.is_tracked or force_save:  # if the portfolio is not "tracked", we do no drift weights
                leftover_positions_ids = list(
                    self.portfolio.assets.filter(date__in=dates).values_list("id", flat=True)
                )  # we need to get the ids otherwise the queryset is reevaluated later
                logger.info(f"bulk saving {len(positions)} positions ({len(leftover_positions_ids)} leftovers) ...")
                objs = AssetPosition.unannotated_objects.bulk_create(
                    positions,
                    update_fields=[
                        "weighting",
                        "initial_price",
                        "initial_currency_fx_rate",
                        "initial_shares",
                        "currency_fx_rate_instrument_to_usd",
                        "currency_fx_rate_portfolio_to_usd",
                        "underlying_quote_price",
                        "portfolio",
                        "portfolio_created",
                        "underlying_instrument",
                    ],
                    unique_fields=["portfolio", "date", "underlying_quote", "portfolio_created"],
                    update_conflicts=True,
                    batch_size=10000,
                )
                if delete_leftovers:
                    objs_ids = list(map(lambda x: x.id, objs))
                    leftover_positions_ids = list(filter(lambda i: i not in objs_ids, leftover_positions_ids))
                    logger.info(f"deleting {len(leftover_positions_ids)} leftover positions..")
                    AssetPosition.objects.filter(id__in=leftover_positions_ids).delete()

            for val_date in self.get_dates():
                if self.portfolio.is_tracked:
                    try:
                        changed_portfolio = self._get_portfolio(val_date)
                    except KeyError:
                        changed_portfolio = None
                    self._change_at_date_tasks[val_date] = changed_portfolio
                    self._compute_metrics_tasks.add(val_date)
        self._positions = defaultdict(dict)

    def schedule_metric_computation(self):
        if self._compute_metrics_tasks:
            basket_id = self.portfolio.id
            basket_content_type_id = ContentType.objects.get_by_natural_key("wbportfolio", "portfolio").id
            group(
                *[
                    compute_metrics_as_task.si(d, basket_id=basket_id, basket_content_type_id=basket_content_type_id)
                    for d in self._compute_metrics_tasks
                ]
            ).apply_async()
            self._change_at_date_tasks = dict()

    def schedule_change_at_dates(self, synchronous: bool = True, **task_kwargs):
        from wbportfolio.models.portfolio import trigger_portfolio_change_as_task

        if self._change_at_date_tasks:
            tasks = chain(
                *[
                    trigger_portfolio_change_as_task.si(
                        self.portfolio.id, d, changed_portfolio=portfolio, evaluate_rebalancer=False, **task_kwargs
                    )
                    for d, portfolio in self._change_at_date_tasks.items()
                ]
            )
            if synchronous:
                tasks.apply()
            else:
                tasks.apply_async()
            self._change_at_date_tasks = dict()



from django.db.models import Q, TextChoices
from django.utils.timezone import localdate
from wbcore import filters as wb_filters
from wbcore.pandas.filterset import PandasFilterSetMixin
from wbcore.utils.date import current_financial_month
from wbfdm.filters.utils import last_period_date_range
from wbfdm.models import Instrument

from wbportfolio.models import Product

from .products import BaseProductFilterSet


def current_date(*args, **kwargs):
    return localdate()


class PerformancePandasFilter(PandasFilterSetMixin, BaseProductFilterSet):
    class PerformanceBy(TextChoices):
        NNM = "NNM", "NNM"
        AUM = "AUM", "AUM"
        PRICE = "PRICE", "Price"

    is_forex_fix = wb_filters.BooleanFilter(label="Fix Forex rate", method="fake_filter", initial=False, required=True)

    date = wb_filters.FinancialPerformanceDateRangeFilter(
        label="Date Range",
        method=lambda queryset, label, value: queryset,
        required=True,
        clearable=False,
        initial=current_financial_month,
    )

    performance_by = wb_filters.ChoiceFilter(
        label="Performance By",
        choices=PerformanceBy.choices,
        initial=PerformanceBy.PRICE.name,
        method="fake_filter",
        clearable=False,
        required=True,
    )
    white_label_customers = classifications = classifications_neq = unclassified = invested = portfolio = (
        content_type
    ) = tags = None

    class Meta:
        model = Product
        fields = {}
        df_fields = {
            "diff__gte": wb_filters.NumberFilter(
                precision=2, label="Difference", lookup_expr="gte", field_name="diff"
            ),
            "perf__gte": wb_filters.NumberFilter(
                precision=2, label="Performance", lookup_expr="gte", field_name="perf"
            ),
            "n1__gte": wb_filters.NumberFilter(precision=2, label="N1", lookup_expr="gte", field_name="n1"),
            "n2__gte": wb_filters.NumberFilter(precision=2, label="N2", lookup_expr="gte", field_name="n2"),
            "n1_usd__gte": wb_filters.NumberFilter(
                precision=2, label="N1 (USD)", lookup_expr="gte", field_name="n1_usd"
            ),
            "n2_usd__gte": wb_filters.NumberFilter(
                precision=2, label="N2 (USD)", lookup_expr="gte", field_name="n2_usd"
            ),
            "diff_usd__gte": wb_filters.NumberFilter(
                precision=2, label="Difference (USD)", lookup_expr="gte", field_name="diff_usd"
            ),
            "perf_usd__gte": wb_filters.NumberFilter(
                precision=2, label="Performance (USD)", lookup_expr="gte", field_name="perf_usd"
            ),
            "diff__lte": wb_filters.NumberFilter(
                precision=2, label="Difference", lookup_expr="lte", field_name="diff"
            ),
            "perf__lte": wb_filters.NumberFilter(
                precision=2, label="Performance", lookup_expr="lte", field_name="perf"
            ),
            "n1__lte": wb_filters.NumberFilter(precision=2, label="N1", lookup_expr="lte", field_name="n1"),
            "n2__lte": wb_filters.NumberFilter(precision=2, label="N2", lookup_expr="lte", field_name="n2"),
            "n1_usd__lte": wb_filters.NumberFilter(
                precision=2, label="N1 (USD)", lookup_expr="lte", field_name="n1_usd"
            ),
            "n2_usd__lte": wb_filters.NumberFilter(
                precision=2, label="N2 (USD)", lookup_expr="lte", field_name="n2_usd"
            ),
            "diff_usd__lte": wb_filters.NumberFilter(
                precision=2, label="Difference (USD)", lookup_expr="lte", field_name="diff_usd"
            ),
            "perf_usd__lte": wb_filters.NumberFilter(
                precision=2, label="Performance (USD)", lookup_expr="lte", field_name="perf_usd"
            ),
        }


class ProductPerformanceNetNewMoneyFilter(PandasFilterSetMixin, BaseProductFilterSet):
    is_active = wb_filters.BooleanFilter(label="Is Active", method="filter_active_products", initial=True)
    net_negative_money__gte = wb_filters.NumberFilter(
        label="Net Negative money", lookup_expr="gte", field_name="net_negative_money"
    )
    net_negative_money__lte = wb_filters.NumberFilter(
        label="Net Negative money", lookup_expr="lte", field_name="net_negative_money"
    )
    net_positive_money__gte = wb_filters.NumberFilter(
        label="Net Positive money", lookup_expr="gte", field_name="net_positive_money"
    )
    net_positive_money__lte = wb_filters.NumberFilter(
        label="Net Positive money", lookup_expr="lte", field_name="net_positive_money"
    )
    net_positive_money_usd__gte = wb_filters.NumberFilter(
        label="Net Positive money ($)", lookup_expr="gte", field_name="net_positive_money_usd"
    )
    net_positive_money_usd__lte = wb_filters.NumberFilter(
        label="Net Positive money ($)", lookup_expr="lte", field_name="net_positive_money_usd"
    )
    net_negative_money_usd__gte = wb_filters.NumberFilter(
        label="Net Negative money ($)", lookup_expr="gte", field_name="net_negative_money_usd"
    )
    net_negative_money_usd__lte = wb_filters.NumberFilter(
        label="Net Negative money ($)", lookup_expr="lte", field_name="net_negative_money_usd"
    )
    net_money__gte = wb_filters.NumberFilter(label="Net Money", lookup_expr="gte", field_name="net_money")
    net_money__lte = wb_filters.NumberFilter(label="Net Money", lookup_expr="lte", field_name="net_money")
    net_money_usd__gte = wb_filters.NumberFilter(label="Net Money ($)", lookup_expr="gte", field_name="net_money_usd")
    net_money_usd__lte = wb_filters.NumberFilter(label="Net Money ($)", lookup_expr="lte", field_name="net_money_usd")

    date = wb_filters.DateRangeFilter(
        label="Date Range",
        method=lambda queryset, label, value: queryset,
        required=True,
        clearable=False,
        initial=last_period_date_range,
    )

    def filter_active_products(self, queryset, name, value):
        if value is True:
            return queryset.filter(~Q(net_money=0.0))
        elif value is False:
            return queryset.filter(Q(net_money=0.0))

        return queryset

    white_label_customers = classifications = classifications_neq = unclassified = invested = portfolio = (
        content_type
    ) = tags = None

    class Meta:
        model = Product
        fields = {
            "computed_str": ["exact", "icontains"],
        }


class PerformanceComparisonFilter(PandasFilterSetMixin, BaseProductFilterSet):
    classifications = unclassified = invested = None
    dates = wb_filters.FinancialPerformanceDateRangeFilter(
        method=lambda q, n, v: q, label="Compute Performance between these dates"
    )

    comparison_instrument = wb_filters.ModelChoiceFilter(
        label="Compare with...",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
        filter_params={"level": 0},
        method="fake_filter",
    )
    compare_primary_benchmark = wb_filters.BooleanFilter(
        initial=False, label="Compare against primary benchmark", method="fake_filter"
    )

    class Meta:
        model = Product
        fields = {}



from decimal import Decimal
from unittest.mock import patch

import pytest
from faker import Faker
from pandas.tseries.offsets import BDay

from wbportfolio.factories import AdjustmentFactory
from wbportfolio.models import Adjustment, PortfolioRole
from wbportfolio.models.adjustments import post_adjustment_on_prices

fake = Faker()


@pytest.mark.django_db
class TestAdjustmentModel:
    @pytest.fixture()
    def old_adjustment(self):
        return AdjustmentFactory.create(status=Adjustment.Status.PENDING, date=fake.past_date())

    # Test utilities function
    @patch("wbportfolio.models.adjustments.chain")
    def test_adjustment_creation(
        self,
        mock_fct,
        user_factory,
        adjustment_factory,
        product_portfolio_role_factory,
        asset_position_factory,
        instrument,
    ):
        user_factory.create()  # User without permission
        user_porftolio_manager = user_factory.create()
        product_portfolio_role_factory.create(
            person=user_porftolio_manager.profile, role_type=PortfolioRole.RoleType.PORTFOLIO_MANAGER
        )
        asset_position_factory.create(underlying_instrument=instrument)
        adjustment2 = adjustment_factory.create(instrument=instrument, date=fake.past_date())
        assert mock_fct.call_count == 1
        assert adjustment2.status == Adjustment.Status.PENDING

        # If newly created instrument's adjustment happens in the future, its state becomes pending and an import task is triggered
        adjustment1 = adjustment_factory.create(instrument=instrument)
        assert adjustment1.status == Adjustment.Status.PENDING
        assert mock_fct.call_count == 2

    def test_cumulative_factor(self, adjustment_factory):
        """
        s1 -> s2 -> s3
        """
        s3 = adjustment_factory.create(status=Adjustment.Status.APPLIED)
        s3.refresh_from_db()
        assert s3.cumulative_factor == Decimal(1.0)

        s2 = adjustment_factory.create(
            status=Adjustment.Status.APPLIED, instrument=s3.instrument, date=s3.date - BDay(1)
        )
        s1 = adjustment_factory.create(
            instrument=s3.instrument, date=s3.date - BDay(2)
        )  # This adjustment is not applied and shouldn't be accounted for in the cumulative factor of the futures adjustments
        s3.refresh_from_db()
        s2.refresh_from_db()
        s1.refresh_from_db()
        assert s2.cumulative_factor == pytest.approx(
            s3.factor, rel=Decimal(1e-3)
        )  # Normal case, applied adjustment cumulative factor should reflect the parent factors
        assert s1.cumulative_factor == pytest.approx(
            s2.factor * s3.factor, rel=Decimal(1e-3)
        )  # even if not applied, cumulative factor is computed normally

    def test_apply_adjustment_on_assets(self, adjustment, asset_position_factory):
        a1 = asset_position_factory.create(
            underlying_instrument=adjustment.instrument, date=adjustment.adjustment_date
        )
        a2 = asset_position_factory.create(underlying_instrument=adjustment.instrument, date=adjustment.date)
        adjustment.apply_adjustment_on_assets()
        a1.refresh_from_db()
        a2.refresh_from_db()
        assert a1.applied_adjustment == adjustment
        assert not a2.applied_adjustment

    def test_future_adjustment_not_applied_on_assets(self, adjustment, asset_position_factory):
        p = asset_position_factory.create(
            underlying_instrument=adjustment.instrument, date=(adjustment.date + BDay(1)).date()
        )

        adjustment.apply_adjustment_on_assets()
        p.refresh_from_db()
        assert not p.applied_adjustment

    def test_adjustment_played_out_of_order(self, adjustment_factory, asset_position_factory):
        s1 = adjustment_factory.create()
        s2 = adjustment_factory.create(instrument=s1.instrument, date=(s1.date + BDay(2)).date())

        a0 = asset_position_factory.create(underlying_instrument=s1.instrument, date=s2.adjustment_date - BDay(3))
        a1 = asset_position_factory.create(
            underlying_instrument=s1.instrument, date=s2.adjustment_date - BDay(2), applied_adjustment=s2
        )
        a2 = asset_position_factory.create(
            underlying_instrument=s1.instrument, date=s2.adjustment_date - BDay(1), applied_adjustment=s2
        )
        a3 = asset_position_factory.create(
            underlying_instrument=s1.instrument, date=s2.adjustment_date, applied_adjustment=s2
        )

        s1.apply_adjustment_on_assets()
        a0.refresh_from_db()
        a1.refresh_from_db()
        a2.refresh_from_db()
        a3.refresh_from_db()

        assert a0.applied_adjustment == s1
        assert a1.applied_adjustment == s1
        assert a2.applied_adjustment == s2
        assert a3.applied_adjustment == s2

    # Test Basic FSM workflow
    @patch("wbportfolio.models.adjustments.apply_adjustment_on_assets_as_task.delay")
    def test_apply(self, mock_fct, old_adjustment):
        assert mock_fct.call_count == 0
        old_adjustment.apply()
        old_adjustment.save()
        assert old_adjustment.status == Adjustment.Status.APPLIED
        mock_fct.assert_called()

    def test_deny(self, old_adjustment, user):
        assert old_adjustment.status == Adjustment.Status.PENDING
        old_adjustment.deny(by=user)
        old_adjustment.save()
        assert old_adjustment.last_handler == user.profile
        assert old_adjustment.status == Adjustment.Status.DENIED

    @patch("wbportfolio.models.adjustments.revert_adjustment_on_assets_as_task.delay")
    def test_revert(self, mock_fct, adjustment_factory):
        adjustment = adjustment_factory.create(status=Adjustment.Status.APPLIED)
        assert mock_fct.call_count == 0
        adjustment.revert()
        adjustment.save()
        assert adjustment.status == Adjustment.Status.PENDING
        mock_fct.assert_called_with(adjustment.id)

    def test_revert_adjustment_on_assets(self, adjustment_factory, asset_position_factory):
        s1 = adjustment_factory.create(status=Adjustment.Status.APPLIED)
        s2 = adjustment_factory.create(
            date=s1.date + BDay(1), instrument=s1.instrument, status=Adjustment.Status.APPLIED
        )
        s3 = adjustment_factory.create(
            date=s2.date + BDay(1), instrument=s1.instrument, status=Adjustment.Status.APPLIED
        )

        a1 = asset_position_factory.create(
            underlying_instrument=s1.instrument, date=s1.adjustment_date, applied_adjustment=s1
        )
        a2 = asset_position_factory.create(
            underlying_instrument=s1.instrument, date=s2.adjustment_date, applied_adjustment=s2
        )
        a3 = asset_position_factory.create(
            underlying_instrument=s1.instrument, date=s3.adjustment_date, applied_adjustment=s3
        )

        s1.revert_adjustment_on_assets()
        a1.refresh_from_db()
        a2.refresh_from_db()
        a3.refresh_from_db()
        assert a1.applied_adjustment == s2
        assert a2.applied_adjustment == s2
        assert a3.applied_adjustment == s3

        s2.revert_adjustment_on_assets()
        a1.refresh_from_db()
        a2.refresh_from_db()
        a3.refresh_from_db()
        assert a1.applied_adjustment == s3
        assert a2.applied_adjustment == s3
        assert a3.applied_adjustment == s3

        s3.revert_adjustment_on_assets()
        a1.refresh_from_db()
        a2.refresh_from_db()
        a3.refresh_from_db()
        assert not a1.applied_adjustment
        assert not a2.applied_adjustment
        assert not a3.applied_adjustment

    def test_post_adjustment_on_prices_without_assets(self, adjustment):
        post_adjustment_on_prices(adjustment.id)
        adjustment.refresh_from_db()
        assert adjustment.status == Adjustment.Status.APPLIED

    @patch.object(Adjustment, "automatically_applied_adjustments_on_assets")
    def test_post_adjustment_automatically_approve_on_prices_with_assets(
        self,
        mock_check_fct,
        adjustment,
        asset_position_factory,
    ):
        a1 = asset_position_factory.create(
            underlying_instrument=adjustment.instrument, date=adjustment.adjustment_date
        )
        mock_check_fct.return_value = True
        post_adjustment_on_prices(adjustment.id)
        a1.refresh_from_db()
        adjustment.refresh_from_db()
        assert a1.applied_adjustment == adjustment
        assert adjustment.status == Adjustment.Status.APPLIED

    @patch("wbportfolio.models.adjustments.send_notification")
    @patch.object(Adjustment, "automatically_applied_adjustments_on_assets")
    def test_post_adjustment_not_automatically_approve_on_prices_with_assets(
        self,
        mock_check_fct,
        mock_delay_fct,
        adjustment,
        asset_position_factory,
        user_factory,
        product_portfolio_role_factory,
    ):
        asset_position_factory.create(
            underlying_instrument=adjustment.instrument, date=(adjustment.date - BDay(1)).date()
        )
        user_factory.create()  # User without permission
        user_porftolio_manager = user_factory.create()
        product_portfolio_role_factory.create(
            person=user_porftolio_manager.profile, role_type=PortfolioRole.RoleType.PORTFOLIO_MANAGER
        )
        mock_check_fct.return_value = False
        post_adjustment_on_prices(adjustment.id)
        adjustment.refresh_from_db()
        assert adjustment.status == Adjustment.Status.PENDING



from wbportfolio.pms.typing import Portfolio
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Model Portfolio Rebalancing")
class ModelPortfolioRebalancing(AbstractRebalancingModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_date = (
            self.trade_date
            if self.model_portfolio.assets.filter(date=self.trade_date).exists()
            else self.last_effective_date
        )

    @property
    def model_portfolio_rel(self):
        return self.portfolio.dependency_through.filter(type="MODEL").first()

    @property
    def model_portfolio(self):
        return self.model_portfolio_rel.dependency_portfolio if self.model_portfolio_rel else None

    @property
    def assets(self):
        return self.model_portfolio.get_positions(self.value_date) if self.model_portfolio else []

    def is_valid(self) -> bool:
        return len(self.assets) > 0

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        for asset in self.assets:
            asset.date = self.value_date
            asset.asset_valuation_date = self.value_date
            positions.append(asset._build_dto())
        return Portfolio(positions=tuple(positions))



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.directory.viewsets.display.entries import CompanyModelDisplay as BaseCompanyModelDisplay
from wbcore.contrib.directory.viewsets.display.entries import PersonModelDisplay as BasePersonModelDisplay
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import (
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

AUM_TABLE = Section(
    key="aum_table_section",
    collapsible=False,
    title=_("AUM By Product"),
    display=Display(
        pages=[
            Page(
                title=_("AUM By Product"),
                layouts={
                    default(): Layout(
                        grid_template_areas=[["aum_table"]],
                        inlines=[Inline(key="aum_table", endpoint="wbportfolio_aum")],
                    )
                },
            ),
        ]
    ),
)

AUM_FIELDS = Section(
    key="aum_fields_section",
    collapsible=False,
    title=_("AUM"),
    display=Display(
        pages=[
            Page(
                title=_("AUM"),
                layouts={
                    default(): Layout(
                        grid_template_areas=[
                            ["asset_under_management", "assets_under_management_currency", "investment_discretion"],
                            ["invested_assets_under_management_usd", "potential", "."],
                        ],
                        grid_auto_columns="minmax(min-content, 1fr)",
                        grid_auto_rows=Style.MIN_CONTENT,
                    )
                },
            ),
        ]
    ),
)


class CompanyModelDisplay(BaseCompanyModelDisplay):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        list_display = super().get_list_display()
        list_display.fields = (
            *list_display.fields[:5],
            dp.Field(key="invested_assets_under_management_usd", label="AUM Invested"),
            dp.Field(key="asset_under_management", label="AUM"),
            dp.Field(key="potential", label="Potential"),
            *list_display.fields[5:],
        )
        return list_display

    AUM_TABLE = AUM_TABLE
    PORTFOLIO_FIELDS = AUM_FIELDS

    def get_instance_display(self) -> Display:
        asset_allocation_section = Section(
            key="asset_allocation_section",
            collapsible=False,
            title=_("Asset Allocation"),
            display=Display(
                pages=[
                    Page(
                        title=_("Asset Allocation"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["asset_allocation"]],
                                inlines=[Inline(key="asset_allocation", endpoint="asset_allocation_table")],
                            )
                        },
                    ),
                ]
            ),
        )

        geographic_focus_section = Section(
            key="geographic_focus_section",
            collapsible=False,
            title=_("Geographic Focus"),
            display=Display(
                pages=[
                    Page(
                        title=_("Geographic Focus"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["geographic_focus"]],
                                inlines=[Inline(key="geographic_focus", endpoint="geographic_focus_table")],
                            )
                        },
                    ),
                ]
            ),
        )

        instance_display = super().get_instance_display()
        if "pk" in self.view.kwargs:
            instance_display.pages.insert(
                1,
                Page(
                    title=_("Customer Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["asset_allocation_section", "geographic_focus_section"]],
                            sections=[asset_allocation_section, geographic_focus_section],
                        )
                    },
                ),
            )

        return instance_display


class PersonModelDisplay(BasePersonModelDisplay):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        list_display = super().get_list_display()

        list_display.fields = (
            *list_display.fields[:7],
            dp.Field(key="invested_assets_under_management_usd", label="AUM Invested"),
            dp.Field(key="asset_under_management", label="AUM"),
            dp.Field(key="potential", label="Potential"),
            *list_display.fields[7:],
        )
        return list_display

    AUM_TABLE = AUM_TABLE
    PORTFOLIO_FIELDS = AUM_FIELDS


class AssetAllocationDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="percent", label="Percent"),
                dp.Field(key="asset_type", label="Type"),
                dp.Field(key="max_investment", label="Max Investment"),
                dp.Field(key="comment", label="Comment"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["company", "percent", "asset_type", "max_investment"], [repeat_field(4, "comment")]]
        )


class GeographicFocusDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="percent", label="Percent"),
                dp.Field(key="country", label="Location"),
                dp.Field(key="comment", label="Comment"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["company", "percent", "country"], [repeat_field(3, "comment")]])



from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer

from wbportfolio.models import Fees
from wbportfolio.serializers.products import ProductRepresentationSerializer


class FeesRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:fees-detail")

    class Meta:
        model = Fees
        fields = ("transaction_subtype", "total_value", "_detail")


class FeesModelSerializer(wb_serializers.ModelSerializer):
    _product = ProductRepresentationSerializer(source="product")
    _currency = CurrencyRepresentationSerializer(source="currency")
    total_value_fx_portfolio = wb_serializers.DecimalField(
        max_digits=14, decimal_places=2, read_only=True, required=False
    )
    total_value_gross_fx_portfolio = wb_serializers.DecimalField(
        max_digits=14, decimal_places=2, read_only=True, required=False
    )

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        if (view := request.parser_context.get("view")) and view.is_manager and instance.import_source:
            return {"import_source": instance.import_source.file.url}
        return {}

    class Meta:
        model = Fees
        decorators = {
            "total_value": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "total_value_fx_portfolio": wb_serializers.decorator(
                position="left", value="{{_portfolio.currency_symbol}}"
            ),
            "total_value_gross": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "total_value_gross_fx_portfolio": wb_serializers.decorator(
                position="left", value="{{_portfolio.currency_symbol}}"
            ),
        }
        fields = (
            "id",
            "_additional_resources",
            "transaction_subtype",
            "calculated",
            "fee_date",
            "product",
            "currency",
            "_product",
            "_currency",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
        )



import os

import dateparser
from xlrd.xldate import xldate_as_datetime


def get_fields_name(model):
    return [field.name for field in model._meta.get_fields()]


def get_file_extension(file):
    name, extension = os.path.splitext(file.name)
    return extension.lower()


def convert_string_to_number(string):
    if string == "nan" or not string:
        return 0.0
    if type(string) in (int, float):
        return float(string)
    try:
        return float(string.replace(" ", "").replace(",", ""))
    except ValueError:
        return 0.0


def parse_date(date, formats: list | None = None):
    if formats is None:
        formats = []
    if isinstance(date, int) or isinstance(date, float):
        return xldate_as_datetime(int(date), 0).date()
    if isinstance(date, str):
        if format:
            return dateparser.parse(date, date_formats=formats).date()
        else:
            return dateparser.parse(date).date()


def extract_exchange_ticker(bbg_ticker):
    try:
        ticker, exchange = bbg_ticker.split(" ")[0:2]
    except Exception:
        ticker, exchange = bbg_ticker, None
    return ticker, exchange



from django.db import models
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.models import WBModel

from wbportfolio.import_export.handlers.register import RegisterImportHandler


class Register(ImportMixin, WBModel):
    import_export_handler_class = RegisterImportHandler

    class RegisterStatus(models.TextChoices):
        ACTIVE = ("ACTIVE", "Active")
        INACTIVE = ("INACTIVE", "Inactive")
        WARNING = ("WARNING", "Warning")

    class RegisterInvestorType(models.TextChoices):
        BANK = ("BANK", "Bank and Financial Institution")
        NOMINEE = ("NOMINEE", "Nominee")
        GLOBAL = ("GLOBAL", "Global")
        NON_FINANCIAL_ENTITY = ("NON_FINANCIAL_ENTITY", "Non Financial Entities/Corporate")

    register_reference = models.CharField(max_length=255, unique=True)
    register_name_1 = models.CharField(max_length=255)
    register_name_2 = models.CharField(max_length=255, default="")
    global_register_reference = models.CharField(max_length=255, default="")
    external_register_reference = models.CharField(max_length=255, default="")

    custodian_reference = models.CharField(max_length=255)
    custodian_name_1 = models.CharField(max_length=255)
    custodian_name_2 = models.CharField(max_length=255, default="")
    custodian_address = models.TextField(default="", blank=True)
    custodian_postcode = models.CharField(max_length=255, default="")

    custodian_city = models.ForeignKey(
        "geography.Geography",
        related_name="register_custodian_city_registers",
        on_delete=models.PROTECT,
        limit_choices_to={"level": 3},
        null=True,
        blank=True,
    )
    custodian_country = models.ForeignKey(
        "geography.Geography",
        related_name="register_custodian_country_registers",
        on_delete=models.PROTECT,
        limit_choices_to={"level": 1},
        null=True,
        blank=True,
    )

    sales_reference = models.CharField(max_length=255, default="")
    dealer_reference = models.CharField(max_length=255, default="")

    outlet_reference = models.CharField(max_length=255)
    outlet_name = models.CharField(max_length=255)
    outlet_address = models.TextField(default="", blank=True)
    outlet_postcode = models.CharField(max_length=255, default="")

    outlet_city = models.ForeignKey(
        "geography.Geography",
        related_name="register_outlet_city_registers",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        limit_choices_to={"level": 3},
    )
    outlet_country = models.ForeignKey(
        "geography.Geography",
        related_name="register_outlet_country_registers",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        limit_choices_to={"level": 1},
    )

    citizenship = models.ForeignKey(
        "geography.Geography",
        related_name="register_citizenship_registers",
        on_delete=models.PROTECT,
        limit_choices_to={"level": 1},
        null=True,
        blank=True,
    )
    residence = models.ForeignKey(
        "geography.Geography",
        related_name="register_residence_registers",
        on_delete=models.PROTECT,
        limit_choices_to={"level": 1},
        null=True,
        blank=True,
    )

    investor_type = models.CharField(
        max_length=24, choices=RegisterInvestorType.choices, default=RegisterInvestorType.BANK
    )
    status = models.CharField(max_length=8, choices=RegisterStatus.choices, default=RegisterStatus.ACTIVE)
    status_message = models.TextField(default="", blank=True)

    opened = models.DateField()
    opened_reference_1 = models.CharField(max_length=255, default="")
    opened_reference_2 = models.CharField(max_length=255, default="")

    updated_reference_1 = models.CharField(max_length=255, default="")
    updated_reference_2 = models.CharField(max_length=255, default="")

    computed_str = models.CharField(max_length=255, null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.register_reference}: {self.register_name_1} {self.register_name_2}"

    class Meta:
        verbose_name = "Register"
        verbose_name_plural = "Registers"

    def save(self, *args, **kwargs):
        self.computed_str = str(self)
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:register"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:registerrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"
